from lolpop.component.metadata_tracker.base_metadata_tracker import BaseMetadataTracker
from lolpop.utils import common_utils as utils
from lolpop.utils import mlflow_utils
#import your libraries here
import mlflow
from mlflow.tracking import MlflowClient as client
from pathlib import Path



@utils.decorate_all_methods([utils.error_handler, 
                             utils.log_execution(), 
                             mlflow_utils.check_active_mlflow_run(mlflow)])
class MLFlowMetadataTracker(BaseMetadataTracker):
    #Override required or default configurations here for your class
    ##Add required configuration here
    __REQUIRED_CONF__ = {
        "config": ["mlflow_tracking_uri", "mlflow_experiment_name"]
    }
    ##Add default configuration here
    #__DEFAULT_CONF__ = {
    #    "config": {}
    #}

    def __init__(self, *args, **kwargs):
        #set normal config
        super().__init__(*args, **kwargs)

        tracking_uri = self._get_config("mlflow_tracking_uri")
        experiment_name = self._get_config("mlflow_experiment_name")

        self.client, self.run = mlflow_utils.connect(tracking_uri, experiment_name)
        self.url = tracking_uri

        self.log("Using MLFlow in experiment %s with run id: %s" %(experiment_name, self.run.info.run_id), level="INFO")

    def log_artifact(self, resource, id, path, *args, **kwargs):
        run = resource[1]
        run_id = run.info.run_id
        resource_id = self.get_resource_id(resource)
        self.client.log_artifact(run_id, path, artifact_path=resource_id)
        self.log("Saving artifact %s to directory %s in artifact directory in run %s" %(path, id, run_id))

    def get_artifact(self, resource, id, *args, **kwargs):
        pass

    def log_tag(self, resource, key, value, *args, **kwargs):
        run = resource[1]
        run_id = run.info.run_id
        resource_id = self.get_resource_id(resource)
        key = "%s.%s" % (resource_id, key)
        self.log("Saving tag key=%s, value=%s to run %s" %(key, value, run_id))
        self.client.set_tag(run_id, key, value)

    def get_tag(self, resource, key, *args, **kwargs):
        run = mlflow_utils.get_run(self.client, resource[1].info.run_id)
        resource_id = self.get_resource_id(resource)
        key = "%s.%s" % (resource_id, key)
        return run.data.tags.get(key)


    def log_metadata(self, resource, id, data, *args, **kwargs):
        run = resource[1]
        run_id = run.info.run_id
        self.log("Saving metadata key=%s, value=%s to run %s" %
                 (id, str(data), run_id))
        #special handling for params, although we still want to save as a tag for retrieval purposes later on
        if "param" in id: 
            for k,v in data.items(): 
                self.client.log_param(run_id, k, v)
        self.log_tag(resource, id, data)


    def get_metadata(self, resource, id, *args, **kwargs):
        value = self.get_tag(resource, id)
        return value

    # id + run_id unique identifies resource
    def create_resource(self, id, type=None, parent=None, *args, **kwargs):
        if type == "experiment": 
            parent_id = self.get_resource_id(parent)
            parent_run = parent[1]
            run = mlflow_utils.create_nested_run(parent_run)
            self.log("Created nested run with id %s under parent %s." %(run.info.run_id, parent_run.info.run_id))
            id = "%s.%s" %(run.info.run_id, parent_id)
        else:
            run = self.run
        return (id, run)

    def get_resource(self, id, type, parent=None, *args, **kwargs):
        run = self.run 
        if type=="experiment": 
            run_id, _ = id.split(".")
            run = self.client.get_run(run_id=run_id)
        else: 
            if parent is not None: 
                run = parent[1]
        return (id, run)

    #can't update runInfo properties like start_time/end_time
    def update_resource(self, resource, updates, *args, **kwargs):
        pass

    def clean_resource(self, resource, type):
        if type=="experiment": 
            #end the nested run, this will make the parent run active againi
            mlflow.end_run()

    def get_prev_resource_version(self, resource, extra_filters=[], num=1, *args, **kwargs):
        run = resource[1]
        id = self.get_resource_id(resource)
        filter_string = ""
        for filter in extra_filters: 
            filter_string = "%s and %s" %(filter_string,filter)
        run_list = self.client.search_runs(
                        experiment_ids=run.info.experiment_id,
                        filter_string="attributes.start_time <= %s and attribute.status = 'FINISHED' and tags.parent_run = 'True' and tags.%s.exists = 'True'%s" %(run.info.start_time, id, filter_string), 
                        order_by=["attributes.start_time DESC"],
                        max_results=num,
                        )
        if len(run_list) > 0: 
            if num == 1:    
                return (resource[0], run_list[0])
            else: 
                return (resource[0], run_list)
        else: 
            return None

    #technically we should probably get model from the model repository, but I don't like metadata_tracker knowing about other components
    def get_currently_deployed_model_version(self, model_version, *args, **kwargs):
        id = self.get_resource_id(model_version)
        deployed_mv = self.get_prev_resource_version(model_version, extra_filters=["tags.%s.status = 'DEPLOYED'" %id])
        #if no deployed model, just get the latest promoted model
        if deployed_mv is None: 
            deployed_mv = self.get_prev_resource_version(
                model_version, extra_filters=["tags.%s.status = 'PROMOTED'" % id])
        #if no deployed model, just get the previous model version 
        if deployed_mv is None: 
            deployed_mv = deployed_mv = self.get_prev_resource_version(model_version)
        return deployed_mv

    def get_prediction_job_model_version(self, prediction_job, *args, **kwargs):
        run = prediction_job[1]
        id = prediction_job[0]
        if "_predictions" in prediction_job:
            id = id[:-12] 

        return (id, run)

    def get_latest_model_resource(self, model, type, *args, **kwargs):
        if type == "prediction_job": 
            id = self.get_resource_id(model)
            prediction_id = id + "_predictions"
            prev_run = self.get_prev_resource_version(model, extra_filters=["tags.%s.exists = 'True'" %prediction_id])
            return (prediction_id, prev_run)
        else: 
            pass 

    def get_winning_experiment(self, model_version, *args, **kwargs):
        winning_experiment_id = self.get_tag(model_version, "winning_experiment_id")
        experiment = self.get_resource(winning_experiment_id, type="experiment", parent=model_version)
        return experiment

    def build_model_lineage(self, model_version, dataset_versions, *args, **kwargs):
        if len(dataset_versions) > 0:
                dataset_arr = [self.get_resource_id(dv) for dv in dataset_versions]
                self.log_tag(model_version, "dataset_versions", dataset_arr)
                self.log_tag(model_version, "dataset_versions.exists", "True")
        else: #if we're training w/o generating data (for some reason?) then just try to get data from the last model version
            model_verison_id = self.get_resource_id(model_version)
            prev_model_version = self.get_prev_resource_version(model_version, extra_filters=[
                                                                "tags.%s.dataset_versions.exists = 'True'" % model_verison_id])
            if prev_model_version is not None: 
                dataset_verisons = self.get_metadata(prev_model_version, id="dataset_versions")
                if dataset_verisons is not None: 
                    self.log_tag(model_version, "dataset_versions",dataset_verisons)
                    #self.log_tag(model_version, "dataset_versions.exists", "True")
            else: 
                self.log("Unable to find previous model version with lineage for model : %s" %self.get_resource_id(model_version))


    #resource is tuple (resource, run)
    def get_resource_id(self, resource, *args, **kwargs):
        resource_out = resource[0]    

        if "/" in resource_out:
            #use file name 
            resource_out = str(Path(resource_out).stem).replace(".", "_")
        
        #run_id = resource[1].info.run_id
        #resource_out = "%s_%s" %(resource_out, run_id)

        return resource_out

    def get_parent_id(self, resource, type=None, *args, **kwargs):
        id = resource[0]
        if type == "experiment": 
            _, model_id = id.split(".")
            return model_id 
        else: #there's no real heirarchy other than this in the mlflow implementation
            return None 
        
    def register_vc_resource(self, resource, vc_info, key=None, additional_metadata={}, *args, **kwargs):
        #check if a git commit is present. if so then we know we did an external save
        if vc_info and "hexsha" in vc_info.keys():
            uri = vc_info.get("uri")
            hexsha = vc_info.get("hexsha")

            id = self.get_resource_id(resource)
            if "." not in id: #don't log experiment existance
                self.log_tag(resource, "exists", "True")
            hexsha_key="hexsha"
            if key is not None: 
                hexsha_key = "%s_hexsha" %key
            uri_key="uri"
            if key is not None: 
                uri_key = "%s_uri" %key
            self.log_tag(resource, hexsha_key, hexsha)
            self.log_tag(resource, uri_key, uri)
            #log this so we can search on existance of resource in past runs 

        if len(additional_metadata) > 0:
            for k, v in additional_metadata.items():
                self.log_metadata(resource, id=k, data=v)

    def get_vc_info(self, resource, key="hexsha", *args, **kwargs):
        vc_info = {"hexsha": self.get_tag(resource, key)}
        return vc_info 

    def log_data_profile(self, resource, file_path, *args, **kwargs):
        id = self.get_resource_id(resource)
        id = id + "_data_profile"
        self.log_artifact(resource, id, file_path)
        
    def get_data_profile(self, resource, id, *args, **kwargs):
        pass

    def log_checks(self, resource, file_path, type="data", *args, **kwargs):
        id = self.get_resource_id(resource)
        id = id + "_%s_checks_report" %type
        self.log_artifact(resource, id, file_path)

    def get_data_checks(self, resource, id, *args, **kwargs):
        pass

    def log_data_comparison(self, resource, file_path, *args, **kwargs):
        id = self.get_resource_id(resource)
        id = id + "_data_comparison"
        self.log_artifact(resource, id, file_path)

    def stop(self):
        mlflow_utils.stop_run(self.run.info.run_id, self.run.info.experiment_id)

    #should this live in model_repository?
    def load_model(self, model_obj, model_version, ref_model, pipeline_config={}, *args, **kwargs):
        model_trainer = self.get_metadata(model_version, "winning_experiment_model_trainer")
        model_cl = utils.load_class(model_trainer)
        dependent_components = {"logger": self.logger, "notifier": self.notifier,  "metadata_tracker": self.metadata_tracker,
                                "metrics_tracker": self.metrics_tracker, "resource_version_control": self.resource_version_control}
        #might want to find a better way to do this next step.
        #it would be nice if you could reconstruct a model trainer class only w/ a model version object
        if ref_model is not None:
            config = ref_model.config
            parent_process = ref_model.parent_process
            params = ref_model.params
            pipeline_conf = ref_model.pipeline_conf
        else:
            config = {}
            parent_process = self.parent_process
            params = {}
            pipeline_conf = {}
        if len(pipeline_config) > 0:
            pipeline_conf = pipeline_config 
        
        model = model_cl(conf=config, pipeline_conf=pipeline_conf, runner_conf=self.runner_conf, parent_process=parent_process,
                         problem_type=self.problem_type, params=params, components=dependent_components)
        #if you passed in a model_obj, we assume you have a pre-trained model object you wish to use
        if model_obj is not None:
            model._set_model(model_obj)
        return model



from lolpop.component.base_component import BaseComponent

class BaseMetricsTracker(BaseComponent): 

    def log_metric(self, id, value, time, *args, **kwargs): 
        pass 

    def get_metric(self, id, time, *args, **kwargs):
        pass

    def log_metrics(self, id, time, *args, **kwargs):
        pass

    def copy_metrics(self, id, time, *args, **kwargs):
        pass

    def log_prediction_metrics(self, id, time, *args, **kwargs):
        pass



from lolpop.component.data_checker.base_data_checker import BaseDataChecker
from lolpop.utils import common_utils as utils
import stumpy
import numpy as np 
from matplotlib import pyplot as plt 
from matplotlib.patches import Rectangle

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class StumpyMatrixProfiler(BaseDataChecker):

    __REQUIRED_CONF__ = {
        "config": ["local_dir"]
    }

    def check_data(self, data, *args, **kwargs):
        """Generates a data check report using Deepchecks.

        Args:
            data (pd.DataFrame): A dataframe of the data to check

        Returns:
            data_report (object): Python object of the data report.
            file_path (string):  Path to the exported report. 
            checks_status (string): Status of the checks ("PASS"/"WARN"/"ERROR", etc.)
        """
        model_target = self._get_config("MODEL_TARGET")

        windows = self._get_config("stumpy_window_size", [30])
        num_discords = self._get_config("stumpy_num_discords", 3)

        #set up plot
        fig, axs = plt.subplots(len(windows)+1, sharex=True, gridspec_kw={'hspace': 0})
        plt.suptitle('Discord (Anomaly/Novelty) Discovery', fontsize='30')
        axs[0].plot(data[model_target].values)
        axs[0].set_ylabel('TS', fontsize='10')
        box_height = int(data[model_target].max())

        j=1
        for m in windows: 
            mp = stumpy.stump(data[model_target], m)
            discords = []
            for i in range(1, num_discords+1):
                discord_idx = np.argsort(mp[:, 0])[-i]
                nearest_neighbor_distance = mp[discord_idx, 0]
                discords.append(discord_idx)
                self.log("Found possible anomaly at index %s, which is %s units from its nearest neighbor" %(discord_idx, nearest_neighbor_distance))    
            self._plot_mp(axs, m, box_height, mp, discords, j)
            j=j+1


        file_path = "%s/STUMP_DISCORD_ANALSIS.PNG" % self._get_config("local_dir")
        plt.savefig(file_path)

        return None, file_path, "PASS"

    def _plot_mp(self, axs, m, h, mp, discords, i): 
        axs[i].set_ylabel('MP (m=%s)' %m, fontsize='10')
        axs[i].set_xlabel('Time', fontsize='10')
        axs[i].plot(mp[:, 0])
         
        for discord_idx in discords: 
            if i == 1:  # only add rectangles for the first window, otherwise it's going to get cluttered.
                rect = Rectangle((discord_idx, 0), m, h, facecolor='lightgrey')
                axs[0].add_patch(rect)
            axs[i].axvline(x=discord_idx, linestyle="dashed")



from lolpop.component.notifier.base_notifier import BaseNotifier


class SlackNotifier(BaseNotifier): 
    __REQUIRED_CONF__ = {
        "config" : ["gcloud_credentials_file"]
    }


    def notify(self, msg, level="ERROR", **kwargs): 
        pass
    # import os
    # from slack_sdk import WebClient
    # from slack_sdk.errors import SlackApiError

    # # Your Slack API token
    # SLACK_API_TOKEN = "xoxb-your-token"

    # # Slack channel to send the message to
    # SLACK_CHANNEL = "#your-channel"

    # # The message to be sent
    # SLACK_MESSAGE = "Your message"

    # client = WebClient(token=SLACK_API_TOKEN)

    # try:
    #     response = client.chat_postMessage(
    #         channel=SLACK_CHANNEL,
    #         text=SLACK_MESSAGE
    #     )
    #     print(response)
    # except SlackApiError as e:
    #     print("Error sending message: {}".format(e))


from lolpop.component import BaseComponent


class Base{{cookiecutter.ComponentType}}(BaseComponent):
    #Add required configuration here
    __REQUIRED_CONF__ = {
        "config": []
    }

    #Add default configuration here 
    __DEFAULT_CONF__ = {
        "config": {}
    }

    def __init__(self, conf, pipeline_conf, runner_conf, **kwargs):
        self.__file_path__ = __file__
        #set normal config
        super().__init__(conf, pipeline_conf, runner_conf, **kwargs)

    # Write your functions here. The Base Class should primarily just implement an interface for 
    # other classes to implement.   
    def my_function(self, required_arg1, required_arg2, *args, **kwargs):
        pass

    def my_other_function(self, required_arg1, required_arg2, *args, **kwargs):
        pass



from lolpop.component.data_connector.base_data_connector import BaseDataConnector
from lolpop.utils import common_utils as utils

import pandas as pd

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class LocalDataConnector(BaseDataConnector):

    def get_data(self, source_path, *args, **kwargs):
        file_type = source_path.split(".")[-1]
        if file_type == "csv":
            data = pd.read_csv(source_path, engine="pyarrow", **kwargs)
        elif file_type == "parquet" or file_type == "pq":
            data = pd.read_parquet(source_path, engine="pyarrow", **kwargs)
        elif file_type == "orc":
            data = pd.read_orc(source_path, engine="pyarrow", **kwargs)
        else: 
            self.log("Unsupported file type provided: %s" %source_path)
        self.log("Successfully loaded data from %s into DataFrame." %source_path)

        return data 

    def save_data(self, data, target_path, *args, **kwargs):
        file_type = target_path.split(".")[-1]
        if file_type == "csv":
            data = data.to_csv(target_path, **kwargs)
        elif file_type == "parquet" or file_type == "pq":
            data = data.to_parquet(target_path, **kwargs)
        self.log("Successfully saved data to %s." %target_path)
        return data



from lolpop.pipeline.base_pipeline import BasePipeline

class BaseDeploy(BasePipeline): 
    def __init__(self,pipeline_type="deploy", *args, **kwargs):
        super().__init__(pipeline_type=pipeline_type, *args, **kwargs)

    def promote_model(self, model_version, *args, **kwargs): 
        pass 

    def deploy_model(self, model_version, *args, **kwaargs): 
        pass 

    def approve_model(self, model_version, *args, **kwargs): 
        pass 


from lolpop.component.model_trainer.base_model_trainer import BaseModelTrainer
from lolpop.utils import common_utils as utils
from xgboost import XGBClassifier, XGBRegressor 

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class XGBoostModelTrainer(BaseModelTrainer): 

    mlflow_module = "xgboost"

    #should set self.model in init
    def __init__(self, problem_type=None, params={}, *args, **kwargs): 
        #set normal config
        super().__init__(problem_type = problem_type, params=params, *args, **kwargs)
 
        #set up model based on problem_type
        if self.problem_type == "classification": 
            self.model = XGBClassifier(**params)
        elif self.problem_type == "regression": 
            self.model = XGBRegressor(**params)
        else: 
            msg = "Unsupported problem type (%s) in trainer %s" %(problem_type, self.name)
            self.notify(msg)
            raise Exception(msg)


    def fit(self, data, *args, **kwargs): 
        self.log("Starting model training with parameters: %s" %str(self.params))
        self.model.fit(data["X_train"], data["y_train"])
        self.log("Finished model training.")
        return self.model  

    def predict(self, data, *args, **kwargs):
        predictions = {}

        predictions["train"] = self.model.predict(data["X_train"])
        if self.problem_type == "classification": 
            predictions["train_proba"] = self.model.predict_proba(data["X_train"])

        if data.get("X_valid") is not None: 
            predictions["valid"] = self.model.predict(data["X_valid"])
            if self.problem_type == "classification": 
                predictions["valid_proba"] = self.model.predict_proba(data["X_valid"])

        if data.get("X_test") is not None: 
            predictions["test"] = self.model.predict(data["X_test"])
            if self.problem_type == "classification": 
                predictions["test_proba"] = self.model.predict_proba(data["X_test"])

        #self.predictions = predictions

        return predictions 

    #def save(self, experiment, *args, **kwargs): 
    #    pass

    def predict_df(self,df): 
        return self.model.predict(df)

    def predict_proba_df(self,df, to_list=False): 
        predictions = self.model.predict_proba(df)
        if to_list: 
            predictions = predictions.tolist()
        return predictions


def __map_runners__():
    from pathlib import Path
    import os
    from importlib import import_module
    from inspect import isclass
    from .base_runner import BaseRunner
    import warnings

    warnings.filterwarnings("ignore")
    #get current directory and all subdirectors. These represent the resource types
    path = Path(__file__).parent.resolve()
    subdirs = ["%s/%s" % (path, x) for x in os.listdir(path)
               if ((x[0] != "_") and (x[0] != ".") and (".py" not in x))]
    #for each resource type (i.e. subdir), get all resources implemented in that type (i.e. python files in the subdir)
    for subdir in subdirs: 
        files = [x for x in os.listdir(subdir) if (
            (".py" in x) and (x[0] != ".")  and ("__" not in x))]
        #from each file, import all classes and register them in the global namespace. 
        for file in files: 
            try: 
                module = import_module("lolpop.%s.%s.%s" %(subdir.split("/")[-2],subdir.split("/")[-1],file[:-3]))
                classes = [x for x in dir(module) if isclass(getattr(module,x))]
                runners = [x for x in classes if issubclass(getattr(module, x), BaseRunner)]
                globals().update({name: getattr(module,name) for name in runners})
            except: 
                pass

    warnings.resetwarnings()

__map_runners__()


from lolpop.component.base_component import BaseComponent

class BaseModelBiasChecker(BaseComponent): 
    __REQUIRED_CONF__ = {
        "config" : []
    }
    
    def check_model_bias(self, data, model, *args, **kwargs): 
        pass 




from lolpop.component.base_component import BaseComponent

class BaseResourceVersionControl(BaseComponent): 

    def version_data(self, id, data, *args, **kwargs): 
        pass 

    def get_data(self, id, *args, **kwargs):
        pass

    def version_model(self, id, model, *args, **kwargs): 
        pass
    
    def get_model(self, id, *args, **kwargs): 
        pass



from lolpop.pipeline.predict.base_predict import BasePredict
from lolpop.utils import common_utils as utils
from lolpop.utils import metaflow_utils as meta_utils
from metaflow import FlowSpec, step
from pathlib import Path

METAFLOW_CLASS = "MetaflowOfflinePredictSpec"
PLUGIN_PATHS = "plugin_paths.txt"

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class MetaflowOfflinePredict(BasePredict):
    __REQUIRED_CONF__ = {
        "components": ["data_connector", "metadata_tracker", "resource_version_control", "model_explainer", "data_checker", "data_profiler"],
        "config": []
    }

    def run(self, model, model_version, data, dataset_version, **kwargs):
        #get flow class object from this file
        mod_cl = meta_utils.get_flow_class(__file__, METAFLOW_CLASS)

        flow = meta_utils.load_flow(
            mod_cl, self, PLUGIN_PATHS, model=model, model_version=model_version, 
            data=data, dataset_version=dataset_version)
        self.log("Loaded metaflow flow %s" % METAFLOW_CLASS)

        meta_utils.run_flow(flow, "run", __file__, PLUGIN_PATHS)
        self.log("Metaflow pipeline %s finished." % METAFLOW_CLASS)

    def get_artifacts(self, artifact_keys):
        #get latest run of this pipeline
        run = meta_utils.get_latest_run(METAFLOW_CLASS)

        #get requested artifacts
        artifacts = meta_utils.get_run_artifacts(
            run, artifact_keys, METAFLOW_CLASS)

        return artifacts


class MetaflowOfflinePredictSpec(FlowSpec):

    def __init__(self, lolpop=None, model=None, model_version=None, data = None, dataset_version = None, use_cli=False, **kwargs):
        #you need to set local attributes before calling super
        #only the first time we create the class will these parameters be provided.
        #The rest of the calls are metaflow internal calls and we do not want to reset them
        if lolpop is not None:
            self.lolpop = lolpop
        if model is not None:
            self.model = model
        if model_version is not None:
            self.model_version = model_version
        if data is not None: 
            self.data = data 
        if dataset_version is not None: 
            self.dataset_version = dataset_version

        #load plugins
        meta_utils.load_plugins(PLUGIN_PATHS)

        #call init -- this kicks off the main workflow
        super().__init__(use_cli=use_cli)

    @step
    def start(self):
        self.next(self.compare_data)

    @step
    def compare_data(self):
        #get training dataset version and df
        vc_info = self.lolpop.metadata_tracker.get_vc_info(
            self.model_version, key="X_train_hexsha")
        train_df = self.lolpop.resource_version_control.get_data(
            self.model_version, vc_info, key="X_train")

        #compare current dataset version with previous dataset version
        comparison_report, file_path = self.lolpop.data_profiler.compare_data(
            self.data, train_df)

        self.lolpop.metadata_tracker.log_data_comparison(
            self.dataset_version,
            file_path=file_path,
            report=comparison_report,
            profiler_class=self.lolpop.data_profiler.name
        )

        self.next(self.get_predictions)

    @step
    def get_predictions(self):
        #get prediction job
        prediction_id = "%s_predictions" % self.lolpop.metadata_tracker.get_resource_id(
            self.model_version)
        prediction_job = self.lolpop.metadata_tracker.create_resource(
            id=prediction_id, type="prediction_job", parent=self.model_version, prediction_count=self.data.shape[0])

        data = self.data
        #drop any metadata columns
        df = data.drop(self.lolpop._get_config("DROP_COLUMNS", []),
                       axis=1, errors="ignore")

        #make predictions
        data["predictions"] = self.model.predict_df(df)
        if self.lolpop.problem_type == "classification":
            data["predictions_proba"] = self.model.predict_proba_df(
                df, to_list=True)

        #get explanations
        data["explanations"] = self.lolpop.model_explainer.get_explanations(
            df, self.model, self.model_version, "predictions", to_list=True)

        #log predictions
        self.lolpop.metrics_tracker.log_prediction_metrics(
            prediction_job, data["predictions"])

        self.data = data 
        self.prediction_job = prediction_job

        self.next(self.track_predictions)

    @step
    def track_predictions(self):
        #version data
        vc_info = self.lolpop.resource_version_control.version_data(
            self.prediction_job, self.data)

        #register version control metadata w/ metadata tracker
        self.lolpop.metadata_tracker.register_vc_resource(
            self.prediction_job, vc_info, file_type="csv")
        
        self.next(self.analyze_prediction_drift)

    @step
    def analyze_prediction_drift(self):
        #get previous dataset version & dataframe
        prev_dataset_version = self.lolpop.metadata_tracker.get_prev_resource_version(
            self.dataset_version)
        if prev_dataset_version is not None:
            vc_info = self.lolpop.metadata_tracker.get_vc_info(prev_dataset_version)
            prev_data = self.lolpop.resource_version_control.get_data(
                prev_dataset_version, vc_info)

            #compare current dataset version with previous dataset version
            comparison_report, file_path = self.lolpop.data_profiler.compare_data(
                self.data, prev_data)

            self.lolpop.metadata_tracker.log_data_comparison(
                self.prediction_job,
                file_path=file_path,
                report=comparison_report,
                profiler_class=self.lolpop.data_profiler.name
            )

        self.next(self.check_predictions)

    @step
    def check_predictions(self):
        #run data checks
        data = self.data.drop(
            ["explanations", "predictions_proba"], axis=1, errors="ignore")
        data_report, file_path, checks_status = self.lolpop.data_checker.check_data(
            data)

        #log data report to metadata tracker
        self.lolpop.metadata_tracker.log_checks(
            self.prediction_job,
            file_path=file_path,
            report=data_report,
            checker_class=self.lolpop.data_checker.name,
            type="prediction"
        )

        if checks_status == "ERROR" or checks_status == "WARN":
            url = self.lolpop.metadata_tracker.url
            self.lolpop.notify(
                "Issues found with data checks. Visit %s for more information." % url, checks_status)
            
        self.next(self.save_predictions)

    @step
    def save_predictions(self):
        table = self.lolpop._get_config("prediction_data")
        self.lolpop.data_connector.save_data(self.data, table)

        self.next(self.end)

    @step
    def end(self): 
        pass


if __name__ == '__main__':
    #need to use CLI here because metaflow calls back into this function to
    #execute each individual step
    MetaflowOfflinePredictSpec(use_cli=True)



from lolpop.component.base_component import BaseComponent

class BaseModelRepository(BaseComponent): 

    def register_model(self, model, *args, **kwargs): 
        pass 

    def promote_model(self, model, *args, **kwargs): 
        pass 

    def approve_model(self, model, *args, **kwargs): 
        pass 


from lolpop.component.data_profiler.base_data_profiler import BaseDataProfiler
from lolpop.utils import common_utils as utils

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class ContinualDataProfiler(BaseDataProfiler): 

    def profile_data(self, data, entry_names=["train"], datetime_columns=[], index_column=None, time_index_column=None, **kwargs): 
        #bit of a hack as this function currently requires an index
        if index_column is None: 
            index_column = self._get_config("model_index") or "ID"
            if "ID" not in data.columns: 
                data["ID"]=data.index
            
        if time_index_column is None: 
            time_index_column = self._get_config("model_time_index") or None 

        if len(datetime_columns) == 0: 
            datetime_columns = self._get_config("model_datetime_columns") or None 
            if datetime_columns is not None: 
                datetime_columns = datetime_columns.split(",")
            else: 
                datetime_columns = []

        data_profile_dict = {
            "dataframes" : [data],
            "entry_names" : entry_names, 
            "datetime_columns" : datetime_columns, 
            "index_column" : index_column, 
            "time_index_column" : time_index_column,
        }
        
        return data_profile_dict, None

    def compare_data(self, data, prev_data, **kwargs): 
        # not implemented
        pass 



from lolpop.pipeline.base_pipeline import BasePipeline

class BasePredict(BasePipeline): 
    def __init__(self, pipeline_type="predict", *args, **kwargs):
        super().__init__(pipeline_type=pipeline_type, *args, **kwargs)



from lolpop.component.data_transformer.base_data_transformer import BaseDataTransformer
from lolpop.utils import common_utils as utils
from lolpop.component.data_connector.bigquery_data_connector import BigQueryDataConnector

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class BigQueryDataTransformer(BaseDataTransformer): 

    def __init__(self, *args, **kwargs): 
        super().__init__(*args, **kwargs)
        self.bigquery_config = utils.load_config(
            ["GOOGLE_KEYFILE", "GOOGLE_PROJECT", "GOOGLE_DATASET"], self.config)

    def transform(self, sql, *args, **kwargs):
        data = BigQueryDataConnector._load_data(sql, self.bigquery_config)
        return data 



from lolpop.component.data_transformer.base_data_transformer import BaseDataTransformer
from lolpop.utils import common_utils as utils
from lolpop.component.data_connector.redshift_data_connector import RedshiftDataConnector


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class RedshiftDataTransformer(BaseDataTransformer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.redshift_config = utils.load_config([
            "REDSHIFT_HOST",
            "REDSHIFT_PORT",
            "REDSHIFT_USER",
            "REDSHIFT_PASSWORD",
            "REDSHIFT_DBNAME",
            "REDSHIFT_SCHEMA",
        ], self.config)
        self.pg_config = {
            k.replace("REDSHIFT", "POSTGRES"): v for k, v in self.redshift_config.items()}

    def transform(self, sql, *args, **kwargs):
        data = RedshiftDataConnector._load_data(
            sql, self.pg_config)
        return data



from lolpop.component.data_connector.postgres_data_connector import PostgresDataConnector
from lolpop.utils import common_utils as utils
@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class RedshiftDataConnector(PostgresDataConnector):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.redshift_config = utils.load_config(
            ["REDSHIFT_HOST", "REDSHIFT_PORT", "REDSHIFT_USER", "REDSHIFT_PASSWORD", "REDSHIFT_DBNAME", "REDSHIFT_SCHEMA"], self.config)
        self.pg_config = {k.replace("REDSHIFT", "POSTGRES"):v for k,v in self.redshift_config.items()}



from lolpop.component.model_explainer.base_model_explainer import BaseModelExplainer
from lolpop.utils import common_utils as utils
import alibi.explainers as alibi
import shap 
import os
from matplotlib import pyplot as plt
import numpy as np

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class AlibiModelExplainer(BaseModelExplainer): 
    __REQUIRED_CONF__ = {
        "config" : ["explainer_class", "local_dir"]
    }

    def get_explanations(self, data, model, model_version, label, classification_type=None, to_list=False, skip_explainer_plots=True, *args, **kwargs): 
        #load explainer class and instatiate
        explainer_class = self._get_config("explainer_class")
        explainer_cl = getattr(alibi, explainer_class)
        explainer = explainer_cl(model._get_model(), task=self.problem_type, feature_names=data.columns.to_list())

        #fit the explainer and generate explanations
        explainer.fit()
        explanations = explainer.explain(data)

        #generate shap_plots 
        if not skip_explainer_plots:
            self._get_shap_plots(explanations.shap_values, explanations.expected_value, data, model._get_model(), label, model_version, classification_type)

        #log feature importance         
        self.metadata_tracker.log_metadata(model_version, id = "%s_global_feature_importance" %label, data = explanations.raw.get("importances").get("aggregated"))
  
        if to_list: 
            explanations = explanations.raw.get("instances").tolist() 

        return explanations

    def get_feature_importance(self, data_dict, model, model_version, *args, **kwargs): 
        #generate train/test datasets
        (train_X, train_y), (test_X, test_y) = self.data_splitter.get_train_test_dfs(data_dict, combine_xy=False) 
        if self.problem_type == "classification": 
            classification_type = utils.get_multiclass(train_y.unique())
        else: 
            classification_type = None 

        skip_explainer_plots = self._get_config("skip_explainer_plots", False)
        explanations_train = self.get_explanations(train_X, model, model_version, "train", classification_type, skip_explainer_plots=skip_explainer_plots)
        explanations_test = self.get_explanations(test_X, model, model_version, "test", classification_type, skip_explainer_plots=True)

        #compare test and train 
        expected_value_diff, feature_importance_diff = self._compare_train_test_feat_importance(explanations_train, explanations_test, classification_type)
        
        #log feature importance
        self.metadata_tracker.log_metadata(model_version, id = "feature_importance_expected_value_diff_train_test", data = expected_value_diff)
        self.metadata_tracker.log_metadata(model_version, id = "feature_importance_feature_value_diff_train_test", data = feature_importance_diff)

        return explanations_train, explanations_test

    #generate all the shap plot
    def _get_shap_plots(self, shap_values, expected_value, data, model, label, model_version, classification_type=None): 

        #note: classificaiton_type == None currently when making a prediction
        if classification_type == "multiclass":
            #bar plot w/ all classes
            shap.summary_plot(shap_values, data, show=False)
            self._save_pyplot("shap_summary_plot_bar", label, model_version)
        else:
            expected_value = [expected_value] 

        
        for i in range(len(shap_values)):
            ##Note: for multiclass, many plots only work on one class at a time.
            class_str = "" 
            if classification_type == "multiclass": 
                class_str = "_" + str(i)
            #scatter plot for class
            shap.summary_plot(shap_values[i], data, show=False)
            self._save_pyplot("shap_summary_plot_scatter%s" %class_str, label, model_version)

            #bar plot for class
            shap.summary_plot(shap_values[i], data, plot_type="bar", show=False)
            self._save_pyplot("shap_summary_plot_bar%s" %class_str, label, model_version)

            ##force plot for class
            ##force plot is very slow for large matrices, so we'll impose a sample
            sample_size = 100
            idx = np.random.randint(len(shap_values[i]), size=sample_size)
            # force plot doesn't really work w/ multiclass right now
            if classification_type !="multiclass":
                hplt = shap.force_plot(expected_value[i], shap_values[i][idx,:], data.loc[idx] ,show=False)
                force_html = "<head>%s</head> <body>%s</body>" %(shap.getjs(), hplt.html())
                self._save_file("shap_force_plot_bar%s" %class_str, label, force_html, model_version)

            #decision plots
            shap.decision_plot(expected_value[i], shap_values[i][idx,:], show=False)
            self._save_pyplot("shap_decision_plot%s" %class_str, label, model_version)

            #create dependence plots -- one for each feature
            for ft in data.columns: 
                #you can set interaction_index to determine which feature is used to color plot.
                #if omitted it will just use what seems to have the best interaction. If none it will turn off coloring
                shap.dependence_plot(ft, shap_values[i], data, interaction_index=None, show=False)
                self._save_pyplot("shap_dependence_plot%s_%s" %(class_str, ft), label, model_version)\

                #partial dependenceplots
                if i == 0: #only need to calculate these once
                    shap.partial_dependence_plot(ft, model.predict, data, show=False)
                    self._save_pyplot("shap_partial_dependence_plot_%s" %ft, label, model_version)

            # we can also check bias too if not doing it elsewhere: 
            #for bias in config.get('BIAS_COLUMNS').split(","): 
            #    shap.group_difference_plot(shap_values[i], data[bias].values==1, feature_names=data.columns)
            #    save_pyplot("shap_group_difference_plot_%s_bias_%s" %(str(i),bias), label, model_version)
            
            #waterfall_plot only shows the effect of one row at a time
            #need to construct proper exp[lanation object for this plot instead of using alibi wrapper
            shap_ex = shap._explanation.Explanation(shap_values[i], base_values=expected_value)
            shap.waterfall_plot(shap_ex[0], show=False)
            self._save_pyplot("shap_waterfall_example%s" %class_str, label, model_version)

            #heatmap
            #commenting out because this takes a long time to generate
            #shap.plots.heatmap(shap_ex)
            #self._save_pyplot("shap_heatmap%s" %class_str, label, model_version)
            
            #other plots
            #multioutput_decision_plot only shows the effect of one row at a time, but it's super awesome for multiclass problems
            
   #save shapley plut using matplotlib.pyplot
    def _save_pyplot(self, name, label, model_version): 
        local_path = "%s/%s/%s" %(self._get_config("local_dir"), self.metadata_tracker.get_resource_id(model_version), label)
        os.makedirs(local_path, exist_ok=True)
        key = "%s_%s"%(name, label)
        local_file = "%s/%s.png" %(local_path, key)
        plt.savefig(local_file)
        artifact = self.metadata_tracker.log_artifact(model_version, id = key, path = local_file, external=False)
        #plt.clf()
        plt.close()

        return artifact

    def _save_file(self, name, label, content, model_version, extension="html"): 
        local_path = "%s/%s/%s" % (self._get_config("local_dir"),self.metadata_tracker.get_resource_id(model_version), label)
        os.makedirs(local_path, exist_ok=True)
        key = "%s_%s" % (name, label)
        local_file = "%s/%s.%s" % (local_path, key, extension)
        with open(local_file, "w+") as f: 
            f.write(content)
        artifact = self.metadata_tracker.log_artifact(model_version, id=key, path=local_file, external=False)
 
        return artifact

    def _compare_train_test_feat_importance(self, explanations_train, explanations_test, classification_type, threshold=0.25): 
        #shap_train = explanations_train.shap_values
        #shap_test = explanations_test.shap_values

        #make everything 3dim so the following code is the same regardless of type
        #if classification_type !="multiclass": 
        #    shap_train = [shap_train]
        #    shap_test = [shap_test]

        #check expected values 
        if classification_type == "multiclass": 
            expected_diff = [(x - y)/max(x, 0.00001) for x,y in zip(explanations_train.expected_value, explanations_test.expected_value)]
        else: 
            expected_diff = [(explanations_train.expected_value -
                             explanations_test.expected_value)/max(explanations_train.expected_value, 0.00001)]

        for x in expected_diff: 
            if x> threshold: 
                self.notify("Threshold exceeded in difference between train/test shap expected values.")

        #check avg value for feature
        # we could recalculate everything, but shap provides the globals in the aggregated features
        #np_train = np.array(shap_train)
        #train_ft_values = np.abs(np_train).mean(axis=0).mean(axis=0)
        #np_test = np.array(shap_test)
        #test_ft_values = np.abs(np_test).mean(axis=0).mean(axis=0)

        raw_train_importances = explanations_train.raw.get("importances").get("aggregated")
        raw_test_importances = explanations_test.raw.get("importances").get("aggregated")
        # sort values to ensure features match up
        train_ft_values = sorted(
            [ x[1] for x in  zip(raw_train_importances.get("names"), raw_train_importances.get("ranked_effect"))] 
            ) 
        test_ft_values = sorted(
            [ x[1] for x in  zip(raw_test_importances.get("names"), raw_test_importances.get("ranked_effect"))]
            ) 

        feature_diff = [(x-y)/max(x, 0.00001) for x,y in zip(train_ft_values, test_ft_values)]

        for x in feature_diff: 
            if x > threshold: 
                self.notify("Threshold exceeded in difference between a feature shap value impact in train & test datasets.")

        return expected_diff, feature_diff
 



from lolpop.component.data_transformer.base_data_transformer import BaseDataTransformer
from lolpop.utils import common_utils as utils
from omegaconf import OmegaConf 

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class dbtDataTransformer(BaseDataTransformer): 
    
    #use load_config to allow setting "DBT_TARGET", "DBT_PROFILE", "DBT_PROJECT_DIR", "DBT_PROFILES_DIR",  via env variables
    __REQUIRED_CONF__ = {
        "config" : ["data_connector"]
    }

    def __init__(self, components={}, *args, **kwargs): 
        super().__init__(components=components, *args, **kwargs)

        self.dbt_config = utils.load_config(["DBT_TARGET", "DBT_PROFILE", "DBT_PROJECT_DIR", "DBT_PROFILES_DIR"], self.config)

        data_connector = self._get_config("data_connector")

        #dbt doesn't actually retrieve data, so we have to use another data_transformer class to do that. 
        # we'll read credentials for that in the dbt profile though, so you don't have to additionally include that
        # in your dbt configuration
        if data_connector is not None:
            config = get_dw_config_from_profile(self.dbt_config)
            obj = utils.register_component_class(self, config, "data_connector", data_connector, self.pipeline_conf, self.runner_conf,
                                           parent_process=self.parent_process, problem_type=self.problem_type, dependent_components=components)

    def get_data(self, source_table_name, *args, **kwargs): 
        """Gets Data. Uses the specified data_connector to get the requested table.

        Args:
            source_table_name (String): Name of table to retrieve

        Returns:
            pd.DataFrame: The data
        """
        return self.data_connector.get_data(source_table_name, *args, **kwargs)

    def transform(self, source_table_name, *args, **kwargs):
        """Runs dbt workflow, specifid by dbt configuration provided. 

        Args:
            source_table_name (String): Name of the table to load and return. This should be a
              table created in the dbt workflow.

        Returns:
            pd.DataFrame: the transformed data
        """
        command = ["dbt", "run", 
                   "--target", self.dbt_config.get("DBT_TARGET"), 
                   "--project-dir", self.dbt_config.get("DBT_PROJECT_DIR"), 
                   "--profiles-dir", self.dbt_config.get("DBT_PROFILES_DIR"),
                   "--profile", self.dbt_config.get("DBT_PROFILE")]

        output, exit_code = utils.execute_cmd(command)

        self.log("dbt output: \n%s" %output, "INFO")

        if int(exit_code) == 0: #dbt ran successfully
            data = self.data_connector.get_data(source_table_name)

        return data 

def get_dw_config_from_profile(dbt_config):
    """Retrieves DW configurtion from dbt profile. 

    Args:
        dbt_config (dict): dictionary containing the dbt configuraiton

    Returns:
        dict: The DW configuration
    """
    conf = OmegaConf.load("%s/profiles.yml" %dbt_config.get("DBT_PROFILES_DIR")).get(dbt_config.get("DBT_PROFILE")).get("outputs").get(dbt_config.get("DBT_TARGET"))
    #get backend_type and figure out what keys we need to map into lolpop 
    backend_type = conf.get("type").lower()
    if backend_type == "snowflake": 
        config_keys = ["account", "database","password", "schema", "user", "warehouse"]
    elif backend_type == "duckdb": 
        config_keys = ["path"]
    elif backend_type == "databricks": 
        config_keys = ["catalog", "schema", "host", "http_path", "token"]
    elif backend_type == "bigquery": 
        method = config.get("method")
        if method == "service-account":
            config_keys=["project", "dataset", "keyfile"]
        else: #using oauth, so assume default application credentials are already set
            config_keys = ["project", "dataset"]
    elif backend_type == "redshift": 
        config_keys=["host", "port", "user", "password", "dbname", "schema"]
    elif backend_type == "postgres": 
        config_keys = ["host", "port", "user", "password", "dbname", "schema"]
    else: 
        config_keys = ["account", "database", "password", "schema", "user", "warehouse"]

    #extract keys and map into lolpop 
    config = {"%s_%s" %(backend_type, x.lower()):y 
              for x,y in conf.items() 
              if x.lower() in config_keys}
    config = OmegaConf.create(
        {"components": {}, "data_connector": {"config": config}})
    
    return config 


from lolpop.component.logger.base_logger import BaseLogger
import logging 
from datetime import datetime 
class FileLogger(BaseLogger): 

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        filename = self._get_config("log_filename","lolpop.log")
        self.url = filename 
        level = self._get_config("log_level", "INFO")
        if isinstance(level, str):
            level = self._get_level_value(level)
        format = self._get_config("log_format", "%(message)s")

        logger = logging.getLogger("lolpop")
        logger.setLevel(level)
        file_handler = logging.FileHandler(filename)
        file_handler.setLevel(level)
        file_handler.setFormatter(logging.Formatter(format))
        logger.addHandler(file_handler)

        self.logger = logger
        logger.log(30, "\n\n")
        self.log("Initialized logger for lolpop workflow.", level="INFO")

    def log(self, msg, level="INFO", time = None, process_name=None, line_num=None, *args, **kwargs): 
        if self._get_level_value(level) >= self._get_level_value(self._get_config("log_level", "DEBUG")):
            msg_out = ""
            if not time:
                time = datetime.utcnow().strftime("%Y/%m/%d %H:%M:%S.%f")
            msg_out = msg_out + "%s [%s] " % (time, level)
            if process_name:
                msg_out = msg_out + \
                    "<%s" % (process_name)
                if line_num and self._get_config("use_line_numbers", False):
                    msg_out = msg_out + \
                        "|%s> ::: %s" % (
                            line_num, msg)
                else:
                    msg_out = msg_out + "> ::: %s" % msg
            else: 
                msg_out = msg_out + "::: %s" % msg
            self.logger.log(self._get_level_value(level), msg_out, **kwargs)

    def _get_level_value(self, level):
        if level == "FATAL" or level == "CRITICAL":
            return 50
        elif level == "ERROR":
            return 40
        elif level == "WARN" or level == "WARNING":
            return 30
        elif level == "INFO":
            return 20
        elif level == "DEBUG" or level == "TRACE":
            return 10
        elif level == "ALL":
            return 1
        else: #level == "NONE" or level == "NOTSET":
            return 0



from lolpop.component.notifier.base_notifier import BaseNotifier

#mainly to be used for development/testing purposes... when you don't actually want to notify


class StdOutNotifier(BaseNotifier):
    __REQUIRED_CONF__ = {
        "config": []
    }

    def notify(self, msg, level="ERROR", **kwargs):
        print(msg)


from lolpop.utils import common_utils as utils
from omegaconf import OmegaConf
from inspect import currentframe
class BasePipeline: 

    __REQUIRED_CONF__ = {
        "components": [], 
        "config": []
    }
    __DEFAULT_CONF__ = {
        "config" : {}
    }

    suppress_logger = False
    suppress_notifier = False

    def __init__(self, conf={}, runner_conf={}, parent_process="runner", problem_type=None, 
                 pipeline_type="base_pipeline", components={}, plugin_mods=[], plugin_paths=[], 
                 skip_config_validation=False, *args, **kwargs):
        #set basic properties like configs
        self.name = type(self).__name__
        self.type = self.__module__.split(".")[-2]
        self.integration_type = self.__module__.split(".")[-1]
        conf = utils.get_conf(conf)
        self.parent_process = parent_process
        self.pipeline_type = pipeline_type
        self.runner_conf = runner_conf
        self.problem_type = problem_type

        #resolve any variables
        conf = utils.resolve_conf_variables(conf)
        #merge into default conf
        conf = utils.copy_config_into(OmegaConf.create(conf), self.__DEFAULT_CONF__)
        #merge pipeline conf into runner conf
        valid_conf = conf.copy()
        OmegaConf.update(valid_conf, "config", utils.copy_config_into(valid_conf.get("config", {}), runner_conf))
        #validate configuration
        if not skip_config_validation:
            valid_conf = self._validate_conf(valid_conf, components)
        self.config = conf.get("config", {})

        #handle default components: logger, notifier, metadata_tracker
        components = utils.set_up_default_components(self, valid_conf, self.runner_conf,
                                                            plugin_mods=plugin_mods,
                                                            skip_config_validation=skip_config_validation,
                                                            components=components)

        #set up reference to each component that is passed in from runner. 
        for component in components.keys(): 
            setattr(self, component, components.get(component))

        pipeline_conf = self.config
        pipeline_conf["pipeline_type"]=self.pipeline_type

        #handle plugins
        if len(plugin_mods) == 0: 
            if len(plugin_paths) == 0:
                plugin_paths = self._get_config("plugin_paths", [])
            file_path = None
            if hasattr(self, "__file_path__"):
                file_path = self.__file_path__
            plugin_mods = utils.get_plugin_mods(self, plugin_paths, file_path)
        self.plugin_mods = plugin_mods

        #now handle all components explicitly set for pipeline
        #note: this will override any component name inherited from the runner, which is what we want
        pipeline_components = {}
        if "components" in conf.keys(): 
            for component in conf.get("components",{}).keys(): 
                obj = utils.register_component_class(self, conf, component, pipeline_conf = pipeline_conf, 
                                                     runner_conf = runner_conf, parent_process=self.name, 
                                                     problem_type = self.problem_type, dependent_components=components, 
                                                     plugin_mods=plugin_mods, skip_config_validation=skip_config_validation)
                if obj is not None: 
                    self.log("Loaded class %s into component %s" %(type(getattr(self, component)).__name__, component))
                    pipeline_components[component] = obj
                else: 
                    self.log("Unable to load class for component %s" %
                             component)

        #now update all pipeline scope components to know about the other pipeline components
        for obj in pipeline_components.values(): 
            obj._update_components(components = pipeline_components)

    def _validate_conf(self, conf, components, skip_validation=False):
        missing, total_missing = utils.validate_conf(conf, self.__REQUIRED_CONF__, components)
        if total_missing > 0: 
            #check to see if missing components are provided by runner via kwargs
            if len(missing.get("components")) > 0: 
                for component in missing.get("components"): 
                    if components.get(component, None) is not None: 
                        total_missing = total_missing - 1
            if total_missing > 0:   
                raise Exception ("Missing the following from pipeline configuration: %s" %missing)
        return conf

    def log(self, msg, level="INFO", **kwargs): 
        if not self.suppress_logger: 
            self.logger.log(msg, level, process_name=self.name,
                            line_num=currentframe().f_back.f_lineno, **kwargs)

    def notify(self, msg, level="ERROR"): 
        if not self.suppress_notifier: 
            self.notifier.notify(msg, level)
            self.log("Notification Sent: %s" %msg, level)

    #helper function for lookup up config key in pipeline or runner conf
    def _get_config(self, key, default_value=None): 
        key = key.lower()
        value = utils.lower_conf(self.config).get(key, None)
        if value is None: 
            value = utils.lower_conf(self.runner_conf).get(key, default_value)
        return value 

    def _set_config(self, key, value): 
        key = key.lower()
        self.config[key]=value



from lolpop.component.base_component import BaseComponent


class BaseDataConnector(BaseComponent):

    def get_data(self, source, *args, **kwargs):
        pass

    def save_data(self, data, target, *args, **kwargs):
        pass



from lolpop.pipeline.wrapper.base_wrapper import BaseWrapper
from lolpop.utils import common_utils as utils

from metaflow import Flow

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class MetaflowPipelineWrapper(BaseWrapper):

    def run(self):
        metaflow_path = self._get_config("metaflow_script") 
        result, _ = utils.execute_cmd(["python3", metaflow_path, "run"], self)
        #self.log(result)

        flow_id = self._get_config("flow_id")
        if flow_id: 
            flow = Flow(flow_id)
            if flow.latest_run.successful:
                status = "successful"
            else: 
                status = "unsuccessful"
            self.log("Metaflow pipline %s completed. Status is %s" %(flow.latest_run.pathspec,status))
        



from lolpop.component.model_checker.base_model_checker import BaseModelChecker
from lolpop.utils import common_utils as utils
from evidently.test_suite import TestSuite
from evidently.tests import TestColumnDrift
from evidently.report import Report
from evidently.test_preset import MulticlassClassificationTestPreset, NoTargetPerformanceTestPreset, BinaryClassificationTestPreset, RegressionTestPreset
from evidently.metric_preset import DataDriftPreset, TargetDriftPreset, DataQualityPreset
from evidently import ColumnMapping

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class EvidentlyAIModelChecker(BaseModelChecker): 

    def check_model(self, data_dict, model, **kwargs):
        if self.problem_type == "classification": 
            classification_type = utils.get_multiclass(data_dict["y_train"].unique())

        #set up column mapping 
        column_mapping = ColumnMapping()
        model_target = self._get_config("MODEL_TARGET")
        column_mapping.target = model_target
        column_mapping.prediction = "prediction"

        #set up data + predictions for train/test drift
        df_train, df_test = self.data_splitter.get_train_test_dfs(data_dict) 
        df_train["prediction"] = model.predict_df(df_train.drop([model_target], axis=1))
        df_test["prediction"] = model.predict_df(df_test.drop([model_target], axis=1))

        if self.problem_type == "classification": 
            if classification_type == "multiclass": 
                model_report = TestSuite(tests=[MulticlassClassificationTestPreset(), NoTargetPerformanceTestPreset(), TestColumnDrift(column_name=model_target)])
            else: 
                model_report = TestSuite(tests=[BinaryClassificationTestPreset(), NoTargetPerformanceTestPreset(), TestColumnDrift(column_name=model_target)])
        elif self.problem_type == "regression":
            model_report = TestSuite(tests=[RegressionTestPreset(), NoTargetPerformanceTestPreset(), TestColumnDrift(column_name=model_target)])
        else: 
            self.notify("Unsupported problem type: %s" %self.problem_type)
        model_report.run(current_data=df_test, reference_data=df_train, column_mapping=column_mapping)
        file_path = "%s/EVIDENTLY_MODEL_REPORT.HTML" %self._get_config("local_dir")
        model_report.save_html(file_path)

        summary = model_report.as_dict()["summary"]

        checks_status = "PASS"
        if summary["failed_tests"] > 0: 
            checks_status = "ERROR"
        elif summary["success_tests"] < summary["total_tests"]: 
            checks_status = "WARN"

        return  model_report, file_path, checks_status

    def calculate_model_drift(self, data, current_model, deployed_model):
        #set up dfs
        df_current = data["X_test"].copy() 
        df_deployed = df_current.copy()

        #create column mapping
        column_mapping = ColumnMapping()
        column_mapping.target=self._get_config("model_target")
        column_mapping.target="prediction"
        df_current["prediction"] = current_model.predict_df(df_current)
        df_deployed["prediction"] = deployed_model.predict_df(df_deployed)

        #drift_report = TestSuite(tests=[TestColumnDrift(column_name="prediction")])
        drift_report = Report(metrics = [TargetDriftPreset()])

        drift_report.run(current_data=df_current, reference_data=df_deployed, column_mapping=column_mapping)
        file_path = "%s/EVIDENTLY_MODEL_DRIFT_REPORT.HTML" %self._get_config("local_dir")
        drift_report.save_html(file_path)

        return drift_report, file_path
        


from lolpop.component.base_component import BaseComponent

class BaseTestRecorder(BaseComponent):

    def record_test(self, obj, func, result, msg, *args, **kwargs):
        pass

    def print_report(*args, **kwargs): 
        pass 

    def generate_report(*args, **kwargs): 
        pass 



from lolpop.utils import common_utils as utils
from omegaconf import OmegaConf
from inspect import currentframe

class BaseComponent: 
    __REQUIRED_CONF__ = {
        "config" : []
    }
    __DEFAULT_CONF__ = {
        "config" : {}
    }

    suppress_logger = False
    suppress_notifier = False 

    def __init__(self, conf={}, pipeline_conf={}, runner_conf={}, parent_process=None, problem_type = None, 
                 components = {}, skip_config_validation=False, *args, **kwargs):
        #set basic properties, like name and configs
        self.name = type(self).__name__
        self.type = self.__module__.split(".")[-2]
        self.integration_type = self.__module__.split(".")[-1]
        config = utils.get_conf(conf)
        self.pipeline_conf = pipeline_conf
        self.runner_conf = runner_conf
        self.parent_process = parent_process
        self.problem_type = problem_type

        #resolve variables
        config = utils.resolve_conf_variables(config)
        # copy config into the default config
        omega_conf = utils.copy_config_into(OmegaConf.create(config), self.__DEFAULT_CONF__)
        # set config. This ensures that we pick up any default config as well
        valid_conf = omega_conf.copy()
        # update config.config with runner and pipeline config.
        # Needs to be done in this order to keep config precedence
        OmegaConf.update(valid_conf, "config",
                         utils.copy_config_into(valid_conf.get("config", {}),
                                                utils.copy_config_into(pipeline_conf, runner_conf)
                                                )
                        )

        #Now we validate config to make sure the component has everything it needs
        if not skip_config_validation: 
            self._validate_conf(valid_conf, components)
        self.config = omega_conf.get("config", {})

        #handle default components: logger, notifier, metadata_tracker
        components = utils.set_up_default_components(self, valid_conf, self.runner_conf,
                                                     skip_config_validation=skip_config_validation,
                                                     components=components)

        #if the config looks good, then we can set all our components 
        for component in components.keys(): 
            setattr(self, component, components.get(component))
        
    def _update_components(self, components = {}, *args, **kwargs): 
        for component in components.keys(): 
            setattr(self, component, components.get(component))

    def _validate_conf(self, conf, components):
        missing, total_missing = utils.validate_conf(conf, self.__REQUIRED_CONF__, components)
        if total_missing>0: 
            #check to see if missing components are provided by runner
            if len(missing.get("components",{})) > 0: 
                for component in missing.get("components"): 
                    if components.get(component, None) is not None: 
                        total_missing = total_missing - 1
            if total_missing > 0:   
                raise Exception ("Missing the following from %s component configuration: %s" %(type(self).__name__, missing))

    def log(self, msg, level="INFO", **kwargs): 
        if not self.suppress_logger:
            self.logger.log(msg, level, process_name=self.name, line_num=currentframe().f_back.f_lineno, **kwargs)


    def notify(self, msg, level="ERROR"): 
        if not self.suppress_notifier: 
            self.notifier.notify(msg, level)
            self.log("Notification Sent: %s" %msg, level)

    #helper function to lookup config key in component, pipeline or runner conf
    def _get_config(self, key, default_value=None):
        key = key.lower()
        value = utils.lower_conf(self.config).get(key, None)
        if value is None: 
            value = utils.lower_conf(self.pipeline_conf).get(key, None)
            if value is None: 
                value = utils.lower_conf(self.runner_conf).get(key, default_value)
        return value

    def _set_config(self, key, value): 
        key = key.lower()
        self.config[key]=value



from lolpop.pipeline.base_pipeline import BasePipeline

class BaseProcess(BasePipeline): 

    datasets_used = []

    def __init__(self, pipeline_type = "process", *args, **kwargs):
        super().__init__(pipeline_type=pipeline_type, *args, **kwargs)

    def transform_data(): 
        pass 

    def track_data(): 
        pass

    def profile_data(): 
        pass 

    def check_data(): 
        pass 

    def compare_data(): 
        pass 




from lolpop.pipeline.deploy.base_deploy import BaseDeploy
from lolpop.utils import common_utils as utils


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class OfflineDeploy(BaseDeploy):

    def promote_model(self, model_version, model=None, *args, **kwargs):
        #get model from RVC if not provided.
        if model is None:
            if model_version is not None:
                experiment = self.metadata_tracker.get_winning_experiment(model_version)
                model_obj = self.resource_version_control.get_model(experiment)
                model = self.metadata_tracker.load_model(model_obj, model_version, ref_model=None)
            else: 
                self.notify("Must provide either model_version or model in order to promote a model.")

        #register model in model repository
        model_id = self.model_repository.register_model(model_version, model)

        #promote model
        promotion = self.model_repository.promote_model(model_id)

        return promotion

    def check_approval(self, promotion, *args, **kwargs):
        is_approved = self.model_repository.check_approval(promotion)

        if not is_approved and self._get_config("auto_approve_models"):
            self.approve_model(promotion)
            is_approved = True

        return is_approved

    def approve_model(self, promotion, *args, **kwargs):
        approval = self.model_repository.approve_model(promotion)
        return approval

    def deploy_model(self, promotion, model_version, *args, **kwaargs):
        deployment = self.model_deployer.deploy_model(promotion, model_version)
        return deployment



import typer 
from typer.core import TyperGroup

class NaturalOrderGroup(TyperGroup):
    def list_commands(self, ctx):
        return self.commands.keys()
    
app = typer.Typer(cls = NaturalOrderGroup, help="{{cookiecutter.command_description}}")

@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        typer.echo(ctx.command.get_help(ctx))

@app.command("my_first_command", help="This is the first command in the command group.")
def my_first_command(
    my_first_argument: str = typer.Argument(..., help="A description of this argument"),
    my_first_option: str = typer.Option(None, help="A description of this option."),
    my_second_option: str = typer.Option(None, help="A description of this option."),
): 
    #your code here 
    pass 


@app.command("my_second_command", help="This is the first second in the command group.")
def my_second_command(
    my_first_argument: str = typer.Argument(...,
                                            help="A description of this argument"),
    my_first_option: str = typer.Option(
        None, help="A description of this option."),
    my_second_option: str = typer.Option(
        None, help="A description of this option."),
):
    #your code here
    pass



from lolpop.component.metrics_tracker.base_metrics_tracker import BaseMetricsTracker
from lolpop.component.metadata_tracker.continual_metadata_tracker import ContinualMetadataTracker
from lolpop.utils import continual_utils as cutils
from lolpop.utils import common_utils as utils

#@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class ContinualMetricsTracker(BaseMetricsTracker): 
    __REQUIRED_CONF__ = {
        "config" : ["CONTINUAL_APIKEY", "CONTINUAL_ENDPOINT", "CONTINUAL_PROJECT", "CONTINUAL_ENVIRONMENT"]
    }
    
    def __init__(self, description=None, run_id=None, components = {}, *args, **kwargs): 
        #set normal config
        super().__init__(components = components, *args, **kwargs)
        
        # if we are using continual for metadata tracking then we won't have to set up connection to continual
        # if not, then we do. If would be weird to have to do this, but just in case. 
        if isinstance(components.get("metadata_tracker"), ContinualMetadataTracker): 
            self.client = self.metadata_tracker.client
            self.run = self.metadata_tracker.run
        else: 
            secrets = utils.load_config(["CONTINUAL_APIKEY", "CONTINUAL_ENDPOINT", "CONTINUAL_PROJECT", "CONTINUAL_ENVIRONMENT"], self.config)
            self.client = cutils.get_client(secrets)
            self.run = cutils.get_run(self.client, description=description, run_id=run_id)

    def create_metric(self, resource, id, display_name, **kwargs):
        metric = resource.metrics.create(id=id, display_name=id)
        self.log("Continual metric created: %s" %(metric.name))
        return metric

    def log_metric(self, resource, id, value, group=None, time=None, **kwargs): 
        #first, check to see if metric exists and if not, just create it 
        metric_name = "%s/metrics/%s" %(resource.name,id)
        try: 
            metric = self.get_metric(resource, metric_name)
        except: 
            metric = None 
        if metric is None: 
            metric = self.create_metric(resource, id, id)

        metric_val = metric.log(value=value, group = group, timestamp=time)
        self.log("Created metric value %s/metrics/%s with value %s" %(resource.name, id, value))
        return metric_val
        
    def get_metric(self, resource, id):
        metric = resource.metrics.get(id=id)
        return metric 

    #if latest, returns latest metric, otherwise returns the entire series
    def get_metric_value(self, resource, id, group=None, latest=True):
        if id == "performance_metric_val" or id == "performance_metric_name":
            metric = getattr(resource, id)
            return metric
        else:
            metric = resource.metrics.get(id=id)

            metric_vals = metric.values
            if group is not None:
                metric_vals = [x.value for x in metric_vals if x.group == group]
            if latest:
                metric_vals = metric_vals[-1].value

            return metric_vals

    def log_metrics(self, resource, metrics, perf_metric):
        is_experiment = resource.name.split("/")[-2] == "experiments"

        if is_experiment: 
            #set performance metric for experiment 
            resource.performance_metric_name = perf_metric

            if len(metrics.get("valid")) > 0: 
                resource.performance_metric_val = metrics["valid"][perf_metric]
            else: 
                resource.performance_metric_val = metrics["train"][perf_metric]
            resource.update(paths=["performance_metric_name", "performance_metric_val"])

        #create metrics object for each metric
        metrics_dicts = cutils.convert_to_metrics_dicts(metrics)
        for key in metrics["train"].keys(): 
            self.create_metric(resource, id=key, display_name=key)

        #log all metrics 
        for metric in metrics_dicts: 
            self.log_metric(resource, **metric)
    
    #copy metrics from one resource to another. 
    #Mainly used to copy winning exp metrics to model version
    def copy_metrics(self, from_resource, to_resource, to_type="model_version"): 
        #iterate through all metrics in from_resource. these will have types Metric
        for old_metric in from_resource.metrics.list_all(): 
            #create new metric for each metric
            new_metric = self.create_metric(to_resource,id=old_metric.id, display_name=old_metric.display_name)
            #iterate through all metric vaules in old_metric. These will have type MetricValue
            for value in old_metric.values(): 
                #log a new value for each MetricValue
                new_metric.log( 
                    value = value.value, 
                    group = value.group, 
                    timestamp = value.timestamp, 
                    step = value.step, 
                    label = value.label,
                )
        if to_type == "model_version": 
            to_resource.update(
                performance_metric_name = from_resource.performance_metric_name,
                performance_metric_val = from_resource.performance_metric_val
                )
        elif to_type =="experiment": 
            to_resource.performance_metric_name = from_resource.performance_metric_name
            to_resource.performance_metric_val = from_resource.performance_metric_val
            to_resource.update(paths=["performance_metric_name", "performance_metric_val"])

    def log_prediction_metrics(self, prediction_job, predictions): 
        df_values = predictions.value_counts()
        class_distribution = {x:y/sum(df_values) for x,y in df_values.to_dict().items()}

        self.metadata_tracker.log_metadata(prediction_job, id="class_distribution", data = class_distribution)
        self.log_metric(prediction_job, "num_predictions", len(predictions))


from lolpop.component.base_component import BaseComponent


class BaseGenAIChatbot(BaseComponent):

    def ask(self, prompt, *args, **kwargs):
        pass

    def prepare_message(role, content, *args, **kwargs): 
        pass


from lolpop.component.data_transformer.base_data_transformer import BaseDataTransformer
from lolpop.utils import common_utils as utils
from lolpop.component.data_connector.databricks_sql_data_connector import DatabricksSQLDataConnector


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class DatabricksSQLDataTransformer(BaseDataTransformer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.databricks_config = utils.load_config([
            "DATABRICKS_SERVER_HOST",
            "DATABRICKS_HTTP_PATH",
            "DATABRICKS_TOKEN",
            "DATABRICKS_CATALOG",
            "DATABRICKS_SCHEMA",
            ], self.config)

    def transform(self, sql, *args, **kwargs):
        data = DatabricksSQLDataConnector._load_data(
            sql, self.databricks_config)
        return data



from lolpop.component.base_component import BaseComponent

class BaseModelDeployer(BaseComponent): 

    def deploy_model(self, model, *args, **kwargs): 
        pass 

    def build_api(self, model, *args, **kwargs): 
        pass 

    def build_container(self, model, *args, **kwargs): 
        pass 

    def package_model(self, model, *args, **kwargs): 
        pass


from lolpop.component.model_bias_checker.base_model_bias_checker import BaseModelBiasChecker
from lolpop.utils import common_utils as utils
from aif360.datasets import StandardDataset
from aif360.metrics import BinaryLabelDatasetMetric, ClassificationMetric
from omegaconf.listconfig import ListConfig
@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class AIFairnessModelBiasChecker(BaseModelBiasChecker): 

    def check_model_bias(self, data, model, model_version, *args, **kwargs): 
        if self.problem_type == "classification": 
            #set up data/predictions
            df_train = data["X_train"]
            df_train_w_labels = df_train.copy() 
            df_train_w_labels[self._get_config("model_target")] = data["y_train"].values
            preds = model.predict(data).get("train")

            #set up aif dataset + configs
            model_target = self._get_config("model_target")
            privileged_groups = self._get_config("privileged_groups")
            unprivileged_groups = self._get_config("unprivileged_groups")
            favorable_classes = self._eval_classes(self._get_config("favorable_classes"))
            privileged_classes = self._eval_classes(self._get_config("privileged_classes"))
            ds = StandardDataset(
                df=df_train_w_labels.dropna(), #NA make AI360 fail 
                label_name=model_target, 
                protected_attribute_names=self._get_config("protected_attribute_names"), 
                favorable_classes=favorable_classes, 
                privileged_classes=privileged_classes,
                )

            #get fairness metrics from binarylabeldatasetmetric. This essentially checks bias in the dataset itself
            bldm = BinaryLabelDatasetMetric(ds, unprivileged_groups=unprivileged_groups, privileged_groups=privileged_groups)
            self.metrics_tracker.log_metric(model_version, "bldm_base_rate", bldm.base_rate()) # ratio of priv/total 
            self.metrics_tracker.log_metric(model_version, "bldm_disparate_impact", bldm.disparate_impact()) # ratio of prob fav outcomes of non/priv
            self.metrics_tracker.log_metric(model_version, "bldm_mean_difference",  bldm.mean_difference()) # diff in probability of favorable outcoems for non vs priv
            self.metrics_tracker.log_metric(model_version, "bldm_consistency", bldm.consistency()[0]) # meastures how similar the labels are for similar instanes
            self.metrics_tracker.log_metric(model_version, "bldm_smoothed_edf", bldm.smoothed_empirical_differential_fairness()) #

            ds_preds = ds.copy()
            ds_preds.labels=preds

            #get fairness metric from classificationmetric. This checks fairness in predictions
            cm = ClassificationMetric(ds, ds_preds, unprivileged_groups=unprivileged_groups, privileged_groups=privileged_groups)
            self.metrics_tracker.log_metric(model_version, "cm_tpr_diff", cm.true_positive_rate_difference()) 
            self.metrics_tracker.log_metric(model_version, "cm_selection_rate", cm.selection_rate()) 
            self.metrics_tracker.log_metric(model_version, "cm_theil_index", cm.theil_index()) 
            self.metrics_tracker.log_metric(model_version, "cm_false_positive_rate_ratio", cm.false_positive_rate_ratio()) 
            self.metrics_tracker.log_metric(model_version, "cm_false_omission_rate_ratio", cm.false_omission_rate_ratio()) 
            self.metrics_tracker.log_metric(model_version, "cm_false_negative_rate_ratio", cm.false_negative_rate_ratio())
            self.metrics_tracker.log_metric(model_version, "cm_false_discovery_rate_ratio", cm.false_discovery_rate_ratio()) 
            self.metrics_tracker.log_metric(model_version, "cm_error_rate_ratio", cm.error_rate_ratio())  
            self.metrics_tracker.log_metric(model_version, "cm_differential_fba", cm.differential_fairness_bias_amplification())
        else: 
            self.log("Problem type %s not supported for model bias checker %s" %(self.problem_type, self.name))        

        #TODO: we should have bias thresholds, and if any are exceeded, then we can reweigh the dataset
        #via something like: https://aif360.readthedocs.io/en/stable/modules/generated/aif360.algorithms.preprocessing.Reweighing.html
        #will have to think about how this applies in the workflow though.
        #i.e. we may want to move the dataset bias stuff up in the data_processing step. 

    def _eval_classes(self, class_list): 
        list_out = []
        for cl in class_list:
            if isinstance(cl, str) and  cl.startswith("lambda x"): 
                cl = eval(cl) #we should rework this in the future 
            elif isinstance(cl, list) or isinstance(cl, ListConfig): 
                cl = self._eval_classes(cl)
            list_out.append(cl)  
        return list_out      


from lolpop.component.hyperparameter_tuner.base_hyperparameter_tuner import BaseHyperparameterTuner
from lolpop.utils import common_utils as utils

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class LocalHyperparameterTuner(BaseHyperparameterTuner): 

    def run_experiment(self, data, model_version, *args, **kwargs): 
    # params can call different algos and user hyperparam tuning, etc 
    # first we generate a list of experiments + scores    
        exp_list = {}
        model_list = {}
        training_params = self._get_config("training_params")
        metrics = self._get_config("metrics")
        perf_metric = self._get_config("perf_metric")
        for algo in training_params: 
            grid = self._build_training_grid(training_params[algo])
            for params in grid:
                #train model 
                model, experiment = self.build_model(data, model_version, algo, params)
                #save model
                self.save_model(model, experiment, params, algo)
                #make predictions
                predictions = model.predict(data)
                #calculate metrics
                metrics_val = model.calculate_metrics(data, predictions, metrics)
                #log metrics
                self.metrics_tracker.log_metrics(experiment, metrics_val, perf_metric)
                #build experiment list
                exp_id = self.metadata_tracker.get_resource_id(experiment)
                exp_list[exp_id] = metrics_val["valid"][perf_metric]
                model_list[exp_id] = model

        #save data splits
        for k,v in data.items(): 
            vc_info = self.resource_version_control.version_data(model_version, v, key=k)
            self.metadata_tracker.register_vc_resource(model_version, vc_info, key=k, file_type="csv")

        #now, we determine overall best experiment and save into model_version
        winning_exp_id = self._get_winning_experiment(exp_list, perf_metric, reverse=utils.get_metric_direction(perf_metric))
        winning_exp = self.metadata_tracker.get_resource(winning_exp_id, parent=model_version, type="experiment")
        winning_exp_model_trainer = self.metadata_tracker.get_metadata(winning_exp, id="model_trainer")
        best_model = model_list.get(winning_exp_id)

        #log important stuff to model version
        self.metrics_tracker.copy_metrics(winning_exp, model_version)
        self.metadata_tracker.log_metadata(model_version, id="winning_experiment_id", data={"winning_experiment_id" : winning_exp_id})
        self.metadata_tracker.log_metadata(model_version, id="winning_experiment_model_trainer", data={"winning_experiment_model_trainer" : winning_exp_model_trainer})

        return best_model


from lolpop.component.base_component import BaseComponent

class BaseDataProfiler(BaseComponent): 

    def profile_data(self, data, *args, **kwargs): 
        pass 

    def compare_data(self, data, prev_data, *args, **kwargs): 
        pass



import typer 
from lolpop.utils import common_utils as utils
from pathlib import Path
import os 
from typer.core import TyperGroup


class NaturalOrderGroup(TyperGroup):
    def list_commands(self, ctx):
        return self.commands.keys()
    
app = typer.Typer(cls = NaturalOrderGroup, help="Generate synthetic data from existing data.")


@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        typer.echo(ctx.command.get_help(ctx))

@app.command("create", help="Create a synthetic dataset.")
def create(
    source_file: Path = typer.Argument(
        ..., help="Path to the source file."),
    datagen_class: str = typer.Option("SDVDataSynthesizer", "--datagen-class", "-c", help="Data Synthesizer class name."),
    synthesizer_class: str = typer.Option(
        "SingleTablePreset", "--synthesizer-class", "-s", help="Class name in the data synthesizer to use to build a synthesizer model."),

    num_rows: int = typer.Option(10000, "--num-rows", "-n", help="The number of rows to generate for the synthetic data."),
    output_path: Path = typer.Option(os.getcwd(), "--output-path", "-o", help="The location to save the generated_data"),
    evaluate_fake_data: bool = typer.Option(False, "--evaluate-fake-data", help="Run evaluator on fake data.")
): 
    typer.secho("Loading datagen class %s" %datagen_class, fg="blue")
    datagen_cl = utils.load_class(datagen_class, "component")
    logger_cl = utils.load_class("StdOutLogger", "component")
    data_generator = datagen_cl(components={"logger": logger_cl()})
    data_generator.suppress_logger = True 
    data_generator.suppress_notifier = True 
    typer.secho("Successfully loaded datagen class %s" % datagen_class, fg="green")

    typer.secho("Loading data for synthesizer from %s" % source_file, fg="blue")
    data, metadata = data_generator.load_data(str(source_file))
    typer.secho("Successfully loaded data!", fg="green")

    typer.secho("Building data synthesizer model. This can take awhile ...", fg="blue")
    synthesizer = data_generator.model_data(data, metadata, synthesizer_class)
    typer.secho("Successfully built data synthesizer model!", fg="green")

    typer.secho("Creating synthetic dataset with %s rows" % num_rows, fg="blue")
    fake_data = data_generator.sample_data(synthesizer, num_rows)
    typer.secho("Successfully created synthetic data! Here's a preview of the first 5 rows:", fg="green")
    typer.secho(fake_data.head())

    if evaluate_fake_data: 
        typer.secho("Evaluating synthetic dataset...", fg="blue")
        quality_report, diagnostic_report = data_generator.evaluate_data(data, fake_data, metadata, synthesizer_class)
        typer.secho("Successfully evaluated synthetic data!", fg="green")

    typer.secho("Saving synthetic data to output directory %s" % output_path, fg="blue")   
    (source_file_name, source_file_type) = str(source_file).split("/")[-1].split(".")
    output_file_path = "%s/%s_syn.%s" %(output_path, source_file_name, source_file_type)
    if source_file_type == "csv": 
        fake_data.to_csv(output_file_path, index=False)
    elif source_file_type in ["parquet", "pq"]: 
        fake_data.to_parquet(output_file_path, index=False)
    else: 
        typer.secho("Unsupported file type: %s" %source_file_type, fg="red")
    typer.secho("Successfully saved synthetic data to %s!" %output_file_path, fg="green")
    




from lolpop.component.data_connector.base_data_connector import BaseDataConnector
from lolpop.utils import common_utils as utils

import pandas as pd
from databricks import sql as databricks_sql
from sqlalchemy import create_engine
from tqdm import tqdm

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class DatabricksSQLDataConnector(BaseDataConnector):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.databricks_config = utils.load_config(
            [
                "DATABRICKS_SERVER_HOSTNAME", 
                "DATABRICKS_HTTP_PATH", 
                "DATABRICKS_TOKEN", 
                "DATABRICKS_CATALOG", 
                "DATABRICKS_SCHEMA",
            ]
            , self.config)

    def get_data(self, table, sql=None, *args, **kwargs):
        if sql is None:
            if table is not None:
                sql = "SELECT * FROM %s" % table
            else:
                raise Exception(
                    "Table or SQL command must be provided to DatabricksConnector in order to retrieve data.")

        data = self._load_data(sql, self.databricks_config)

        return data

    def save_data(self, data, table, *args, **kwargs):
        #check if table already exists
        tables = self.get_data(None, sql="SHOW TABLES")

        #if table exists, see if we've added new columns or deleted columns
        table_exists = tables[tables["tableName"] == table].shape[0] > 0
        if table_exists:
           table_schema = self.get_data(table)
           new_features = [x for x in data.columns if x.upper(
           ) not in table_schema.columns.str.upper()]
           deleted_features = [
               x for x in table_schema.columns if x.upper() not in data.columns.str.upper()]

           #if we have new columns, we need to alter the table definiton
           if len(new_features) > 0:
               sql = "ALTER TABLE %s " % (table)
               for col_name in new_features:
                   col_type = data.dtypes[col_name]
                   dbricks_type = self._map_pandas_col_type_to_databrickssql_type(
                       col_type)
                   sql = sql + ("ADD COLUMN %s %s" %
                                (col_name.upper(), dbricks_type))
               self.get_data(None, sql=sql)

           #if we've deleted any features, then just add blank rows to the dataframe
           if len(deleted_features) > 0:
               for feature in deleted_features:
                   data[feature] = None

        #now we can save our data
        self._save_data(data, table, self.databricks_config)

     #load data into df
    @classmethod
    def _load_data(self, sql, config):
        df = pd.DataFrame()
        with databricks_sql.connect(server_hostname=config.get("DATABRICKS_SERVER_HOSTNAME"),
                         http_path=config.get("DATABRICKS_HTTP_PATH"),
                         access_token=config.get("DATABRICKS_TOKEN")) as connection:

            with connection.cursor() as cursor:
                cursor.execute(sql)

                df = pd.DataFrame.from_records(
                    iter(cursor), columns=[x[0] for x in cursor.description], coerce_float=True)

        return df

    def _save_data(self, data, table, config):
        chunksize = 16384 
        catalog = config.get("DATABRICKS_CATALOG", "")
        schema = config.get("DATABRICKS_SCHEMA", "")
        engine = create_engine("databricks://token:%s@%s?http_path=%s&catalog=%s&schema=%s"
                               %(config.get("DATABRICKS_TOKEN"), config.get("DATABRICKS_SERVER_HOSTNAME"), 
                                 config.get("DATABRICKS_HTTP_PATH"), catalog, schema))
        connection = engine.connect()

        with tqdm(total=len(data), desc="Upload to %s.%s.%s" % (catalog, schema, table)) as pbar:
            for i, cdf in enumerate(utils.chunker(data, chunksize)):
                cdf.to_sql(table.replace(" ", "_"), con=engine,
                        index=False, if_exists="append", chunksize=chunksize, method="multi")
                pbar.update(chunksize)
        connection.close()
        engine.dispose()

    def _map_pandas_col_type_to_databrickssql_type(self, col_type):
        """_summary_

        Args:
            col_type (_type_): _description_

        Returns:
            _type_: _description_
        """
        if col_type.kind == 'M':
            # datetime and timestamp columns in pandas are converted to datetime64[ns] dtype, which corresponds to 'TIMESTAMP'
            column_type = 'TIMESTAMP'
        elif col_type.kind == 'm':
            # timedelta columns in pandas are converted to timedelta64[ns] dtype, which corresponds to 'INTERVAL'
            column_type = 'INTERVAL'
        elif col_type.kind == 'b':
            # boolean columns in pandas are converted to 'bool' dtype, which corresponds to 'boolean'
            column_type = 'BOOLEAN'
        elif col_type.kind == 'i' or col_type.kind == "u":
            column_type = 'BIGINT'
        elif col_type.kind == 'f' or col_type.kind == 'c':
            # float columns in pandas are converted to 'float64' dtype, which corresponds to 'DECIMAL'
            column_type = 'DECIMAL'
        else:
            # All other column types in pandas are assumed to be string columns, which correspond to 'STRING'
            column_type = 'STRING'
        return column_type




from lolpop.component.model_repository.base_model_repository import BaseModelRepository
from lolpop.utils import common_utils as utils
from lolpop.component.metadata_tracker.continual_metadata_tracker import ContinualMetadataTracker
from lolpop.utils import continual_utils as cutils

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class ContinualModelRepository(BaseModelRepository):
    __REQUIRED_CONF__ = {
        "components": ["metadata_tracker|ContinualMetadataTracker"],
        "config": []
    }

    def __init__(self, description=None, run_id=None, components={}, *args, **kwargs):
        #set normal config
        super().__init__(components=components, *args, **kwargs)

        # if we are using continual for metadata tracking then we won't have to set up connection to continual
        # if not, then we do. If would be weird to have to do this, but just in case.
        if isinstance(components.get("metadata_tracker"), ContinualMetadataTracker):
            self.client = self.metadata_tracker.client
            self.run = self.metadata_tracker.run
        else:
            secrets = utils.load_config(["CONTINUAL_APIKEY", "CONTINUAL_ENDPOINT",
                                        "CONTINUAL_PROJECT", "CONTINUAL_ENVIRONMENT"], self.config)
            self.client = cutils.get_client(secrets)
            self.run = cutils.get_run(
                self.client, description=description, run_id=run_id)

    def register_model(self, model_version, model, *args, **kwargs):
        #only supports the Continual Metadata Tracker, so model is already in the system.
        return model_version

    def promote_model(self, model_version, reason="UPLIFT", *args, **kwargs):
        improvement_metric = self.metrics_tracker.get_metric_value(
            model_version, "performance_metric_name")
        improvement_metric_value = self.metrics_tracker.get_metric_value(model_version,"performance_metric_val")
        try: 
            improvement_metric_diff = self.metrics_tracker.get_metric_value(model_version, "deployed_model_perf_metric_diff")
        except: #this fails if the model is the first model, as there is no comparison to the previous model 
            improvement_metric_diff = 0 
        base_improvement_metric_value = improvement_metric_value - improvement_metric_diff
        promotion = self.metadata_tracker.create_resource(
            improvement_metric=improvement_metric, 
            improvement_metric_value=improvement_metric_value, 
            base_improvement_metric_value=base_improvement_metric_value, 
            #improvement_metric_diff=improvement_metric_diff,
            id=None, parent=model_version, type="promotion", reason=reason)
        return promotion

    #approvals are not implemented in Continual, so just return True if called.
    def check_approval(self, promotion, *args, **kwargs):
        return True

    def approve_model(self, promotion, *args, **kwargs):
        return True



from lolpop.component.metrics_tracker.base_metrics_tracker import BaseMetricsTracker
from lolpop.utils import common_utils as utils
#import your libraries here
from lolpop.component.metadata_tracker.mlflow_metadata_tracker import MLFlowMetadataTracker
from lolpop.utils import mlflow_utils
import json 
import mlflow

@utils.decorate_all_methods([utils.error_handler, 
                             utils.log_execution(), 
                             mlflow_utils.check_active_mlflow_run(mlflow)])
class MLFlowMetricsTracker(BaseMetricsTracker):
    #Override required or default configurations here for your class
    ##Add required configuration here
    __REQUIRED_CONF__ = {
        "components" : ["metadata_tracker|MLFlowMetadataTracker"],
        #"config": ["mlflow_tracking_uri", "mlflow_experiment_name"]
    }
    ##Add default configuration here
    #__DEFAULT_CONF__ = {
    #    "config": {}
    #}

    def __init__(self, components={}, *args, **kwargs):
        #set normal config
        super().__init__(components=components, *args, **kwargs)

        # if we are using mlflow for metadata tracking then we won't have to set up connection to mlflow
        # if not, then we do. If would be weird to have to do this, but just in case.
        if isinstance(components.get("metadata_tracker"), MLFlowMetadataTracker):
            self.client = self.metadata_tracker.client
            self.run = self.metadata_tracker.run
            self.url = self.metadata_tracker.url
        else: #we should only use this if we allow metrics tracking w/o metadata tracking
            tracking_uri = self._get_config("mlflow_tracking_uri")
            experiment_name = self._get_config("mlflow_experiment_name")

            self.client, self.run = mlflow_utils.connect(tracking_uri, experiment_name)
            self.url = tracking_uri

            self.log("Using MLFlow in experiment %s with run id: %s" %(experiment_name, self.run.info.run_id), level="INFO")

    def log_metric(self, resource, id, value, **kwargs): 
        run_id = resource[1].info.run_id
        resource_id = self.metadata_tracker.get_resource_id(resource)
        id = "%s.%s" %(resource_id, id)
        self.client.log_metric(run_id, id, value, **kwargs)
        self.log("Saving metric=%s, value=%s in run %s" % (id, value, run_id))
        
    def get_metric(self, resource, id):
        run = mlflow_utils.get_run(self.client, resource[1].info.run_id)
        resource_id = self.metadata_tracker.get_resource_id(resource)
        id = "%s.%s" % (resource_id, id)
        metric =  run.data.metrics.get(id)
        return metric 

    def log_metrics(self, resource, metrics, perf_metric):
        for split in metrics.keys(): 
            for metric, val in metrics[split].items(): 
                key = "%s_%s" %(split, metric)
                self.log_metric(resource, key, val)

        self.metadata_tracker.log_metadata(resource, "performance_metric", perf_metric)
        if len(metrics.get("valid")) > 0:
            self.metadata_tracker.log_metadata(resource, "performance_metric_value", metrics["valid"][perf_metric])
        else:
            self.metadata_tracker.log_metadata(resource, "performance_metric_value", metrics["train"][perf_metric])
    
    #copy metrics from one resource to another. 
    #Mainly used to copy winning exp metrics to model version
    def copy_metrics(self, from_resource, to_resource, **kwargs): 

        from_run = mlflow_utils.get_run(self.client, from_resource[1].info.run_id)

        for k,v in from_run.data.metrics.items():
            key = k.split(".")[-1] 
            self.log_metric(to_resource, key, v)

        self.metadata_tracker.log_metadata(
            to_resource,
            "performance_metric", 
            self.metadata_tracker.get_metadata(from_resource, "performance_metric"))
        self.metadata_tracker.log_metadata(
            to_resource,
            "performance_metric_value",
            self.metadata_tracker.get_metadata(from_resource, "performance_metric_value"))
        self.metadata_tracker.log_metadata(
            to_resource, 
            "training_params", 
            json.loads(self.metadata_tracker.get_metadata(from_resource, "training_params").replace("\'","\"")))

    def log_prediction_metrics(self, prediction_job, predictions):
        prediction_id = self.metadata_tracker.get_resource_id(prediction_job)
        if self.problem_type == "classification":  
            df_values = predictions.value_counts()
            class_distribution = {x:y/sum(df_values) for x,y in df_values.to_dict().items()}
            self.metadata_tracker.log_metadata(prediction_job, id="class_distribution", data = class_distribution)

        self.log_metric(
            prediction_job, "num_predictions", len(predictions))



import subprocess 
import os 
import sys
from importlib import import_module, util as import_util 
from git import Repo
from omegaconf import OmegaConf, dictconfig
from datetime import datetime
from functools import wraps 
import pandas as pd 
from pathlib import Path 
import time

#
##should decide if these things go into the common_utils class or just the abstract classes. 
#

#shell command wrapper
def execute_cmd(cmd, logger=None):
    if logger: 
        logger.log("Executing command: `%s`" %(" ".join(cmd)))
    p = subprocess.run(cmd, stdout=subprocess.PIPE, universal_newlines=True, shell=False)
    return (p.stdout, p.returncode)

# pulls keys out of config or gets from env variable
# i.e. either handle secret manager before getting here and pass in 
# or populate via env vars
def load_config(config_keys, conf): 
    config = {}
    for key in config_keys: 
        if key in conf.keys():
            config[key]=conf[key]
        elif key.lower() in conf.keys(): #handle cases where we're looking for env var but it's overridden in conf w/ lowercase 
            config[key] = conf[key.lower()]
        else: 
            config[key] = os.getenv(key)
    return config

#comits a single file to git
def git_commit_file(file_path, repo_path=None, msg="Commiting file from lolpop", push=True, logger=None):
    repo = Repo(repo_path)
    repo.index.add(file_path)
    hexsha = repo.index.commit(msg).hexsha

    if push: 
        origin = repo.remotes[0]
        result = origin.push()
        if logger: 
            logger.log("Committed and pushed file %s. Result: %s" %(file_path, result[0].summary))

    return hexsha

def load_plugins(plugin_paths=[]):
    plugins = []
    for dir in plugin_paths:
        if dir.exists():
            plugin = load_plugin(dir)
            plugins.append(plugin)
    return plugins

def load_plugin(plugin_path, obj=None): 
    if plugin_path.is_dir(): 
        plugin_dir = str(plugin_path)
        plugin_name = plugin_path.name
    elif plugin_path.is_file():
        plugin_dir = str(plugin_path.parent)
        plugin_name = plugin_path.stem
    else: 
        raise Exception("Invalid plugin path: %s. Path is not a file nor a directory." %plugin_path)
    if plugin_dir: 
        sys.path.append(plugin_dir)
    if plugin_name: 
        mod = import_module(plugin_name)
    if obj: 
        plugin_paths = obj._get_config("plugin_paths",[]) 
        plugin_paths.append(plugin_dir)
        obj._set_config("plugin_paths", plugin_paths)
    return mod

#load class object
def load_class_obj(class_name, class_type="component", parent="lolpop"): 
    module = import_module("%s.%s" %(parent, class_type))
    cl = getattr(module, class_name)
    return cl

def load_class(class_name, class_type="component", plugin_mods = [], self_obj=None): 
    try:
        cl = load_class_obj(class_name, class_type)
    except: 
        try: 
            if self_obj: 
                self_obj.log("Unable to find class %s in build-in resources. Searching extensions..." %class_name)
            cl = load_class_obj(class_name, class_type="extension")
            if self_obj:
                self_obj.log("Found class %s in extensions!" %class_name)
        except: 
            if self_obj:
                self_obj.log(
                    "Unable to find class %s in extensions. Searching plugins modules in %s..." %(class_name, str(plugin_mods)))
            cl = load_class_from_plugin(class_name, plugin_mods, class_type)
            if cl is not None: 
                if self_obj: 
                    self_obj.log("Found class %s in plugins!" % class_name)
            else: 
                if self_obj:
                    self_obj.log("Unable to find class %s in plugins!" %class_name)
    return cl 

#load class object from plugins
def load_class_from_plugin(class_name, plugin_mods, class_type="component"):
    #try to load from each plugin
    cl = None 
    for plugin in plugin_mods: 
        try: 
            cl = load_class_obj(class_name, class_type=class_type, parent=plugin)
            break 
        except Exception as e: 
            #if multiple plugins are used, the one we are looking for will not be in most of them, so just ignore any errors
            continue 
    return cl

#loads a module from a file
def load_module_from_file(file_path): 
    spec = import_util.spec_from_file_location(file_path.stem, file_path)
    mod = import_util.module_from_spec(spec)
    sys.modules[spec.name]=mod #need to do this to properly register module
    spec.loader.exec_module(mod)
    return mod 

#register component class as an attribute of the provided object
def register_component_class(self_obj, conf, component_type, default_class_name=None, 
                             pipeline_conf = {}, runner_conf = {}, parent_process = "runner", 
                             problem_type = None, dependent_components = {}, plugin_mods=[], *args, **kwargs): 
    obj = None
    component_class_name = conf.get("components",{}).get(component_type, default_class_name)
    if component_class_name is not None:
        cl = load_class(component_class_name) 
        if cl is not None: 
            obj = cl(conf=conf.get(component_type, {}), pipeline_conf=pipeline_conf, runner_conf=runner_conf,
                     parent_process=parent_process, problem_type=problem_type, components=dependent_components, *args, **kwargs)
            setattr(self_obj, component_type, obj)
    return obj 

#registers pipeline as an attribute of the provided object
def register_pipeline_class(self_obj, conf, pipeline_type, default_class_name=None, runner_conf = {}, 
                            parent_process = "runner", problem_type = None, dependent_components = {}, 
                            plugin_mods=[], *args, **kwargs): 
    obj = None 
    pipeline_class_name = conf.get("pipelines",{}).get(pipeline_type, default_class_name)
    if pipeline_class_name is not None: 
        cl = load_class(pipeline_class_name, class_type="pipeline")
        if cl is not None: 
            obj = cl(conf=conf.get(pipeline_type, {}), runner_conf=runner_conf, parent_process=parent_process,
                     problem_type=problem_type, components=dependent_components, plugin_mods=plugin_mods, *args, **kwargs)
            setattr(self_obj, pipeline_type, obj)
    return obj

def lower_conf(conf): 
    return {k.lower():v for k,v in conf.items()}
    
#Need to revisit. Seems like this is not actually merging on update, it's reassigning. 
#As a result, some default conf can get lost
def copy_config_into(conf, default_conf): 
    #have to use omegaconf so types match
    default_copy = OmegaConf.create(default_conf.copy())
    updated_conf = OmegaConf.merge(default_copy, conf)
    return updated_conf

#validates configuration with the required conf specified
def validate_conf(config, required_conf, components_objs={}):
    conf = config.copy()
    #convert componenets_obj dict into primitives
    components = OmegaConf.create({k:type(v).__name__ for k,v in components_objs.items()})
    #updates conf.components with components. 
    #we want local config values to overwrite glboals, so we do a two step update
    if conf.get("components", None) is None: 
        conf["components"]=components
    else: 
        components.update(conf.get("components",{}))
        conf.get("components",{}).update(components)

    missing = {}
    total_missing = 0
    if required_conf is not None: 
        if conf is None: 
            conf = {}
        #map everything to lowercase keys to avoid capitalization mismatch
        lconf = lower_conf(conf)
        lrequired_conf = lower_conf(required_conf)
        for k,v in lrequired_conf.items(): 
            missing_k = []
            for val in v: 
                #value can be of form "key|values" which specifies the value(s) allowed
                #this is mainly used to specify the classes supported
                #i.e. something like "MetadataTracker|ContinualMetadataTracker"
                if "|" in val: 
                    val_arr = val.split("|")
                    val_key = val_arr[0]
                    val_value = val_arr[1].split(",")
                else: 
                    val_key = val
                    val_value = None
                if lconf.get(k,{}).get(val_key.lower(),None) is None: 
                    missing_k.append(val) 
                elif val_value is not None: 
                    if lconf.get(k,{}).get(val_key.lower()) not in val_value: 
                        missing_k.append(val)
            missing[k] = missing_k
        for v in missing.values(): 
            total_missing = total_missing + len(v)
    return missing, total_missing

#replaces variable values with their actual values. 
#conf allows pointing to other variable via $node1.node2.node3, etc...
def resolve_conf_variables(conf, main_conf=None):
    if main_conf is None: #need to always have the full conf lineage because variables can pont to arbitrary nodes
        main_conf = conf
    if conf is not None and len(conf) > 0: 
        for k,v in conf.items(): 
            if isinstance(v, dictconfig.DictConfig):
                OmegaConf.update(conf, k, resolve_conf_variables(v, main_conf)) 
            elif isinstance(v, str): 
                if v[0] == "$": 
                    resolved_value = get_conf_value(v[1:], main_conf)
                    if resolved_value: 
                        OmegaConf.update(conf,k, resolved_value)
                    else: 
                        raise Exception("Unable to resolve configuration value: %s" %v)
    return conf

# returns configuration from conf object (file location or python dict)
def get_conf(conf_obj): 
    if isinstance(conf_obj, str) or isinstance(conf_obj, Path):
        conf = OmegaConf.load(conf_obj)
    elif type(conf_obj) is dict: 
        conf = OmegaConf.create(conf_obj)
    elif isinstance(conf_obj, dictconfig.DictConfig):
        conf = conf_obj
    else: 
        raise Exception("Invalid configuration. Configuration must be a file or a a dict.")
    
    return conf 

#returns node value in conf specified by var in the form 'node1.node2.node3...'
def get_conf_value(var, conf): 
    var_arr = var.split(".")
    out = conf
    try: 
        for i in var_arr: 
            out = out.get(i)
    except: 
        out = None 
    return out 

def get_plugin_mods(self_obj, plugin_paths=[], file_path=None):
    # if no plugin_dir is provided, then try to use the parent directory.
    # if the parent directory is lolpop, then it is a built-in runner and we can ignore
    if file_path is not None:
        #directory should be something like <module_name>/<runner>/<runner_type>/<runner_class>.py
        plugin_dir = os.path.dirname(os.path.dirname(
            os.path.dirname(os.path.realpath(file_path))))
        plugin_paths.append(plugin_dir)

    plugin_paths = [Path(dir) for dir in plugin_paths if dir != "lolpop"]

    #load up all plugin modules
    plugins = load_plugins(plugin_paths)
    plugin_mods = []
    for plugin in plugins: 
        if plugin is not None: 
            plugin_mods.append(plugin.__name__)
    
    return plugin_mods

#for a given lolpop object, get all plugin paths in children
def get_all_plugin_paths(obj):
    children = [x for x in dir(obj) if not x.startswith("_")]
    plugin_paths=[]

    for child in children: 
        child_obj = getattr(obj, child)
        if hasattr(child_obj, "_get_config"): 
            paths = child_obj._get_config("plugin_paths", None)
            if paths is not None: 
                plugin_paths = plugin_paths + paths
    
    plugin_paths = [Path(path) for path in set(plugin_paths)]

    return plugin_paths

#wraps logging calls around function execution
def log_execution(level="DEBUG", start_msg = None, end_msg = None, timeit=True):
    def log_decorator(func):
        @wraps(func)
        def wrapper(obj, *args, **kwargs):
            start = start_msg
            end = end_msg
            if not start: 
                start = "Starting execution of %s" %(func.__name__)
            if not end: 
                end = "Finished execution of %s" %(func.__name__)
            obj.log(start, level)
            obj.log("args: %s, kwargs: %s" %(args, kwargs), "TRACE")
            start_time = time.process_time()
            result = func(obj, *args, **kwargs)
            end_time = time.process_time()
            if timeit: 
                end = end + ". Completed in %s seconds." %(str(end_time - start_time))
            obj.log(end, level)
            return result
        return wrapper
    return log_decorator

#wraps function around error handling
def error_handler(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            raise Exception(f"An error occurred: {e}")
    return wrapper

# decorate all public methods in a class with decorator
# should only be applied to class definition
# decorators can be a single decorator or list of decorators that will be applied in reverse order
# i.e. the first decorator in the list would be the topmost decorator if listed out
# (this seems like the intuitive thing to do)
def decorate_all_methods(decorators):
    def decorate(cls):
        decorator_list = decorators
        if not isinstance(decorator_list, list):
            decorator_list=[decorator_list]
        for attr in dir(cls):
            if callable(getattr(cls, attr)) and not attr.startswith("_") and attr !="log" and attr !="notify":
                func = getattr(cls,attr)
                for decorator in decorator_list[::-1]: 
                    func = decorator(func)
                setattr(cls, attr, func)
        return cls
    return decorate

#compares two dataframes and tries to convert columns so that match types if they don't otherwise
def compare_data_schemas(obj, data, prev_data):
    #check to see if schemas are equal
    ok = True 
    if not data.equals(prev_data): 
        #if they are not equal, try to set schemas to be the same. 
        ok = False
        try: 
            data = convert_col_types(data, prev_data)
            ok = True
        except Exception as e: 
            obj.notify("Data Comparison failed. New and old dataframes do not have compatible column types. Error: %s" %str(e), "ERROR")
            raise 
    return data, prev_data, ok

def convert_col_types(data, prev_data):
    df_a = data.copy()
    df_b = prev_data.copy()
    for col in df_a.columns:
        if col in df_b.columns: #if col is net new, we can skip
            if df_a[col].dtype != df_b[col].dtype:
                df_a[col] = df_a[col].astype(df_b[col].dtype)
    return df_a

#True = higher is better (descending)
def get_metric_direction(perf_metric): 
    reverse = True 
    if perf_metric in ["mse", "rmse", "mae", "mdae", "mape", "smape", "rmsle", "msle"]: 
        reverse = False 
    return reverse

#labels should be something like data["y_train"].unique()
def get_multiclass(labels): 
    classification_type = "binary"
    num_classes = len(labels)
    if num_classes > 2: 
        classification_type = "multiclass"

    return classification_type


def create_df_from_file(source_file, engine="pyarrow", **kwargs): 
    _, source_file_type = str(source_file).split("/")[-1].split(".")
    data = pd.DataFrame()
    if source_file_type == "csv":
        data = pd.read_csv(source_file, engine=engine, **kwargs)
    elif source_file_type in ["parquet", "pq"]:
        data = pd.read_parquet(source_file, engine=engine, **kwargs)
    else:
        raise Exception("Unsupported file type: %s" % source_file_type)
    
    return data

def chunker(seq, size):
    return (seq[pos:pos + size] for pos in range(0, len(seq), size))


#generates a test plan from a configuration file
def generate_test_plan(config, test_plan={}, parent=None, logger=None, test_recorder=None): 
    config = get_conf(config)
    components = config.pop("components", {})
    #set up logger
    if logger is None: 
        logger_cl = load_class(components.get("test_logger",None))
        if logger_cl is not None: 
            logger_conf = config.pop("test_logger",{})
            logger = logger_cl(conf=logger_conf)

    #set up test_recorder
    if test_recorder is None:
        test_recorder_cl = load_class(components.get("test_recorder", "LocalTestRecorder"))
        if test_recorder_cl is not None:
            test_recorder_conf = config.pop("test_recorder", {})
            test_recorder = test_recorder_cl(conf=test_recorder_conf)

    #extract tests
    test_plan = extract_tests(config, test_plan, parent)
    
    return test_plan, logger, test_recorder


def extract_tests(config, test_plan={}, parent=None):
    config = get_conf(config)

    #extract tests
    test_list = config.pop("tests", {})
    for method in test_list:
        key = method
        if parent is not None:
            key = "%s.%s" % (parent, key)
        test_plan[key] = test_list[method]

    #anything else should be dependents + tests
    for resource in config.keys():
        key = resource
        if parent is not None:
            key = "%s.%s" % (parent, key)
        test_plan = extract_tests(config[resource], test_plan=test_plan,
                                          parent=key)
        
    return test_plan

#apply test plan to the given object
#this replaces the specified function with a wrapper that now includes 
# the pre- and post-hooks defined in the testing config
def apply_test_plan(runner, test_plan, test_recorder, test_logger=None): 

    # if no specific test logger is provided, just use the runner's log method
    if test_logger is None: 
        test_logger = runner

    for method,tests in test_plan.items(): 
        fn, obj = get_method(runner, method)
        prehooks = load_hooks(tests.get("pre-hooks"))
        posthooks = load_hooks(tests.get("post-hooks"))

        if len(prehooks) > 0 or len(posthooks) > 0: 
            runner = set_method(runner, method, test_plan_decorator(
                obj, test_logger, test_recorder, prehooks, posthooks)(fn))
            
    return runner

#load hooks from file. 
# Note: the "test" function in the hook file is the entry point
def load_hooks(hook_list): 
    hook_list_out = []
    if hook_list is not None: 
        hook_list_out = [
            (Path(x.get("hook_file")).stem, 
             getattr(load_plugin(Path(x.get("hook_file"))), x.get("hook_method","test")),
             x.get("hook_on_error", "continue").lower()) 
             for x in hook_list
        ]
    return hook_list_out

#retrieves the method on a runner path. Here 'method' is dot notation i.e. "runner.pipeline.component.method"
def get_method(obj, method):
    method_arr = method.split(".")
    out = obj, None
    try: 
        for i in method_arr:
            parent = obj
            obj = getattr(obj, i)
        out = obj, parent
    except: 
        out = None, None 
    return out 

#sets the method for the object to the specified function
def set_method(obj, method, fn):
    method_arr = method.split(".")

    if len(method_arr) == 1: 
        setattr(obj, method_arr[0], fn)
    else: 
        setattr(obj, method_arr[0], 
                set_method(getattr(obj, method_arr[0]),".".join(method_arr[1:]),fn))
        
    return obj 

#decorator used with a test plan to apply pre- and post-hooks to a function 
def test_plan_decorator(obj, logger, recorder, prehooks, posthooks, level="DEBUG"):
    def test_decorator(func): 
        @wraps(func)
        def wrapper(*args, **kwargs):
            for prehook in prehooks:
                obj.log("Running pre-hook %s.%s" %(prehook[0], prehook[1].__name__), level)
                try: 
                    passed = prehook[1](obj, *args, **kwargs)
                    msg_out = None
                    if isinstance(passed, tuple):
                        msg_out = passed[1]
                        passed = passed[0]
                    recorder.record_test(obj, func, prehook[0], prehook[1], passed, msg_out)
                    if passed: 
                        logger.log("Pre-hook %s.%s passed." %
                                   (prehook[0], prehook[1].__name__), "INFO")
                    else: 
                        logger.log("Pre-hook %s.%s failed" %
                                   (prehook[0], prehook[1].__name__), "WARN")
                    if msg_out is not None: 
                        logger.log("Pre-hook message: %s" %msg_out)
                    obj.log("Pre-hook %s.%s finished." %(prehook[0], prehook[1].__name__), level)
                except Exception as e: 
                    logger.log("Error occured when running pre-hook %s.%s: %s" % (prehook[0], prehook[1].__name__, str(e)))
                
                
            result = func(*args, **kwargs)
            
            for posthook in posthooks:
                obj.log("Running post-hook %s.%s" %
                        (posthook[0], posthook[1].__name__), level)
                try: 
                    passed = posthook[1](obj, result, *args, **kwargs)
                    msg_out = None
                    if isinstance(passed, tuple): 
                        msg_out = passed[1] 
                        passed = passed[0]
                    recorder.record_test(obj, func, posthook[0], posthook[1], passed, msg_out)
                    if passed:
                        logger.log("Post-hook %s.%s passed." %
                                   (posthook[0], posthook[1].__name__), "INFO")
                    else:
                        logger.log("Post-hook %s.%s failed" %
                                   (posthook[0], posthook[1].__name__), "WARN")
                    if msg_out is not None:
                        logger.log("Pre-hook message: %s" % msg_out)
                    obj.log("Post-hook %s.%s finished." %
                            (posthook[0], posthook[1].__name__), level)
                except Exception as e:
                    logger.log(
                        "Error occured when running post-hook %s.%s: %s" % (posthook[0], posthook[1].__name__, str(e)))
            return result
        return wrapper
    return test_decorator


def set_up_default_components(obj, conf, runner_conf,
                              plugin_mods=[],
                              skip_config_validation=False,
                              components={}):
    
    #these are the defaults, so if you're explicitly building one, skip this to avoid
    #recursion error 
    if obj.type != "logger" and obj.type != "notifier" and obj.type !="metadata_tracker":
        #set up logger first because we want to pass that to all children
        #we set this up separately from the other components in case you want access to the logger in the
        #__init__ function
        if "logger" not in components.keys(): #you were passed a logger from somewhere else, skip
            logger_obj = register_component_class(obj, conf, "logger", default_class_name="StdOutLogger",
                                                runner_conf=runner_conf, parent_process = obj.name, 
                                                plugin_mods=plugin_mods, skip_config_validation=skip_config_validation,
                                                )
            if logger_obj is not None:
                components["logger"] = logger_obj
                obj.log("Loaded class %s into component %s" %(obj.logger.name, "logger"))
            else:
                raise Exception("Unable to find logger class.")
        else: 
            setattr(obj, "logger", components["logger"])

        if "notifier" not in components.keys():#you were passed a notifier from somewhere else, skip
            notifier_obj = register_component_class(obj, conf, "notifier", default_class_name = "StdOutNotifier", 
                                                runner_conf=runner_conf, parent_process=obj.name,
                                                problem_type=obj.problem_type, dependent_components=components,
                                                plugin_mods=plugin_mods, skip_config_validation=skip_config_validation)
            if notifier_obj is not None:
                components["notifier"] = notifier_obj
                obj.log("Loaded class %s into component %s" %
                            (obj.notifier.name, "notifier"))
            else:
                obj.log("Unable to load notifier component.")
        else:
            setattr(obj, "notifier", components["notifier"])

        #we also want to special handle the metadata tracker, so we'll set that up first as well and pass
        #it to all children so they have access in __init__.
        #it's unclear why you might not want to use a metadata tracker,
        #but we sould consider this use case in the future
        if "metadata_tracker" not in components.keys() and obj.type !="metadata_tracker": #you were passed a metadata_tracker from somewhere else, skip
            meta_obj = register_component_class(obj, conf, "metadata_tracker", runner_conf=runner_conf, parent_process=obj.name,
                                                        problem_type=obj.problem_type, dependent_components=components,
                                                        plugin_mods=plugin_mods, skip_config_validation=skip_config_validation)
            if meta_obj is not None:
                components["metadata_tracker"] = meta_obj
                obj.log("Loaded class %s into component %s" %
                            (obj.metadata_tracker.name, "metadata_tracker"))
            else:
                #for local dev you may turn off metadata_tracker, so let's not strictly enforce that it exists for now
                obj.log("Unable to load metadata_tracker component.")
        else:
            setattr(obj, "metadata_tracker", components["metadata_tracker"])

    return components




from pkgutil import extend_path

#allows extensions to work when doing local dev and adding them to sys path
__path__ = extend_path(__path__, __name__)

def __get_template_dir__(): 
    import os 

    LOLPOP_DIR = os.path.dirname(os.path.realpath(__file__))
    TEMPLATE_DIR = LOLPOP_DIR + "/templates"

    return TEMPLATE_DIR

__template_path__ = __get_template_dir__()


from lolpop.component.genai_chatbot.base_genai_chatbot import BaseGenAIChatbot
from lolpop.utils import common_utils as utils

import openai

class OpenAIChatbot(BaseGenAIChatbot):

    def __init__(self, *args, **kwargs): 
        super().__init__(*args, **kwargs)
        self.api_key = utils.load_config(["OPENAI_API_KEY"], self.config).get("OPENAI_API_KEY")

    def ask(self, messages=[], model="gpt-3.5-turbo", **kwargs):
    
        if len(messages) == 0: 
            raise Exception("Error: No messages provided to chatbot. Please try again with non-null content.")
        else: 
            completion = openai.ChatCompletion.create(
                model=model,
                messages=messages,
                **kwargs
            )

            return completion.choices[0].message.content
        
    def prepare_message(self, role=None, content=None, *args, **kwargs): 
        return {"role": role, "content": content}



from lolpop.component.base_component import BaseComponent
from omegaconf.listconfig import ListConfig
from lolpop.utils import common_utils as utils
import itertools 

class BaseHyperparameterTuner(BaseComponent): 
    __REQUIRED_CONF__ = {
        "config" : ["training_params", "metrics", "perf_metric"]
    }
    
    def run_experiment(self, data, *args, **kwargs): 
        pass 

    def build_model(self, data, model_version, algo, params, trainer_config={}): 
        #create experiment and log params
        experiment = self.metadata_tracker.create_resource(id=None, type="experiment", parent=model_version)

        #load model trainer and build model
        model_cl = utils.load_class(algo)
        dependent_components = {"logger" : self.logger, "notifier" : self.notifier,  "metadata_tracker" :self.metadata_tracker, "metrics_tracker": self.metrics_tracker, "resource_version_control": self.resource_version_control}
        model = model_cl(conf=trainer_config, pipeline_conf=self.pipeline_conf, runner_conf=self.runner_conf, 
                         parent_process = self.name, problem_type = self.problem_type, params=params, components=dependent_components) 
        model_obj = model.fit(data)

        return model, experiment 

    def save_model(self, model, experiment, params, algo, **kwargs):
        model_obj = model._get_model()
        #save model
        vc_info = self.resource_version_control.version_model(experiment, model_obj, algo=algo)
        experiment_metadata = {
            "training_params" : params,
            "model_trainer" : algo, 
        }
        self.metadata_tracker.register_vc_resource(experiment, vc_info, additional_metadata = experiment_metadata)

        experiment_id = self.metadata_tracker.get_resource_id(experiment, type="experiment")
        model_artifacts = model.get_artifacts(experiment_id) or {}
        for k,v in model_artifacts.items(): 
            self.metadata_tracker.log_artifact(experiment, k, v)

    #builds a grid of all possible parameter combinations gives a params item with a list of values for each param
    def _build_training_grid(self, params): 
        keys, values = zip(*params.items())
        values = [v if isinstance(v, ListConfig) else str(v).split(",") for v in values]
        grid = [dict(zip(keys,v)) for v in itertools.product(*values)]

        return grid

    #determines the winning experiment from the experiment list 
    def _get_winning_experiment(self, exp_list, perf_metric, reverse): 
        # sort list by values. reverse=True is descending, False is ascending (Default)
        new_list = sorted(exp_list.items(), key=lambda x: x[1], reverse=reverse) 
        return new_list[0][0]


from lolpop.component.metadata_tracker.base_metadata_tracker import BaseMetadataTracker
from lolpop.utils import common_utils as utils
from lolpop.utils import continual_utils as cutils
import pandas as pd
import os 

#@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class ContinualMetadataTracker(BaseMetadataTracker): 
    """
    A class for logging metadata and artifacts to Continual. Inherits from BaseMetadataTracker.

    Args:
        conf (dict): Configuration dicitonary for the metadata tracker componenet.
        pipeline_conf (dict): Configuration dictionary for the pipeline.
        runner_conf (dict): Configuration dictionary for the runner.
        description (str, optional): Description for the run. Defaults to None.
        run_id (str, optional): ID of the run. Defaults to None.

    Attributes:
        __REQUIRED_CONF__: List of required configuration for this component. 
        client (): Client object for interacting with Continual.
        run (): Run object representing the current run in Continual.
        url (str): URL for the Continual application.
    """
    __REQUIRED_CONF__ = {
        "config" : ["CONTINUAL_APIKEY", "CONTINUAL_ENDPOINT", "CONTINUAL_PROJECT", "CONTINUAL_ENVIRONMENT"]
    }

    def __init__(self, description=None, run_id=None, *args, **kwargs):
        """
        Initialize the ContinualMetadataTracker object.

        Args:
            conf (dict): Configuration dictionary for this component.
            pipeline_conf (dict): Configuration dictionary for the pipeline.
            runner_conf (dict): Configuration dictionary for the runner.
            description (str, optional): Description for the run. Defaults to None.
            run_id (str, optional): ID of the run. Defaults to None.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        #set normal config
        super().__init__(*args, **kwargs)

        secrets = utils.load_config(["CONTINUAL_APIKEY", "CONTINUAL_ENDPOINT", "CONTINUAL_PROJECT", "CONTINUAL_ENVIRONMENT"], self.config)
        
        self.client = cutils.get_client(secrets)
        
        self.run = cutils.get_run(self.client, description=description, run_id=run_id)

        self.log("Using Continual with run id: %s" %self.run.id, level="INFO")

        self.url = self.run.continual_app_url
    
    def __exit__(self):
        """
        Stop the ContinualMetadataTracker object.
        """
        self.stop()

    def log_artifact(self, resource,  **kwargs): 
        """
        Log an artifact to Continual.

        Args:
            resource: Resource object to log the artifact to. Typically should be an experiment, modelversion, or datasetversion object. 
            **kwargs: Arbitrary keyword arguments.

        Returns:
            artifact: The created Continual artifact object.
        """
        artifact = resource.artifacts.create(replace_if_exists=True, **kwargs)
        self.log("Continual artifact created: %s" %artifact.name)
        return artifact 

    def get_artifact(self, resource, id):
        """
        Retrieve a Continual artifact.

        Args:
            resource: Resource object the artifact belongs to.
            id (str): ID of the artifact to retrieve.

        Returns:
            artifact: The retrieved Continual artifact object.
        """
        artifact = resource.artifacts.get(id=id)
        return artifact 

    def log_metadata(self, resource, **kwargs):
        """
        Log metadata to Continual.

        Args:
            resource: Resource object to log the metadata to.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            metadata: The created Continual metadata object.
        """
        metadata = resource.metadata.create(replace_if_exists=True, **kwargs)
        self.log("Continual metadata created: %s" %metadata.name)
        return metadata 

    def get_metadata(self, resource,id):
        """
        Retrieve metadata from Continual.

        Args:
            resource: Resource object the metadata belongs to.
            id (str): ID of the metadata to retrieve.

        Returns:
            dict: The retrieved metadata dictionary.
        """
        data = None 
        try: 
            metadata = resource.metadata.get(id=id)
        except: 
            metadata = None 
        if metadata is not None: 
            data = metadata.data
        return data

    def log_tag(self, resource, key, value): 
        """
        Log a tag to Continual.

        Args:
            resource: Resource object to log the tag to.
            key (str): The key of the tag.
            value (str): The value of the tag.
        Returns:
            tag: The created tag object
        """
        tag = resource.tags.create(key=key, value=value)
        self.log("Continual tag created: %s" %tag.name)
        return tag
    
    def get_tag(self, resource, key, **kwargs):
        value = None
        tag = resource.tags.get(key=key)
        if tag is not None: 
            value = tag.value
        return value

    def create_resource(self, id, type=None, parent=None, **kwargs): 
        resource = None 
        if type == "dataset_version": 
            try: 
                resource = self.run.datasets.get(id).dataset_versions.create()
            except: 
                resource = self.run.datasets.create(id).dataset_versions.create()
        elif type == "dataset": 
            resource = self.run.datasets.create(id)
        elif type == "model_version":
            try: 
                resource = self.run.models.get(id).model_versions.create() 
            except: 
                resource = self.run.models.create(id).model_versions.create()  
        elif type == "experiment": 
            resource = parent.experiments.create()
        elif type == "promotion":
            resource = parent.promotions.create(
                reason = kwargs.get("reason", "UPLIFT"),
                model_version = parent.name,
                improvement_metric = kwargs.get("improvement_metric"),
                improvement_metric_value = kwargs.get("improvement_metric_value"),
                base_improvement_metric_value = kwargs.get(
                    "base_improvement_metric_value"),
                #improvement_metric_diff = kwargs.get("improvement_metric_diff")
                )
        elif type == "prediction_job": 
            resource = parent.batch_predictions.create(model_version=parent.name, prediction_count = kwargs.get("prediction_count")) 

        self.log("Get/Created Continual resource with name: %s" %resource.name, level="INFO")
        return resource

    def get_resource(self, id, type, parent=None): 
        resource = None 
        if type == "dataset_version": 
            resource = self.run.dataset_versions.get(id)
        elif type == "dataset": 
            resource = self.run.datasets.get(id)
        elif type == "experiment": 
            resource = self.run.model_versions.get(parent.name).experiments.get(id)
        elif type == "model": 
            resource = self.run.models.get(id)
        return resource

    def update_resource(self, resource, value_dict, type=None): 
        for k,v in value_dict.items():
            setattr(resource,k,v)
        resource.update(paths=value_dict.keys())

    def clean_resource(self, resource, type): 
        pass 

    def get_prev_resource_version(self, resource_version): 
        previous_resource_version = None
        name_arr = resource_version.name.split("/")
        # any resource version will have a slug ending in .../parent/parent_id/resource/id
        parent_resource_type = name_arr[-4] 
        parent_id = name_arr[-3]
        parent_resource = getattr(self.run,parent_resource_type).get(parent_id)
        resource_version_type = parent_resource_type[:-1] + "_" + name_arr[-2]
        recent_resource_version_list = getattr(parent_resource,resource_version_type).list(default_sort_order="DESC")

        found_current_version = False 
        i = 0
        while not found_current_version and i < len(recent_resource_version_list): 
            i+=1
            found_current_version = (recent_resource_version_list[i-1].name == resource_version.name)

        if found_current_version: 
            if i == len(recent_resource_version_list): #handle edge case where it was the last version in the list 
                previous_resource_version = getattr(parent_resource,resource_version_type).list(latest=True, page_size = len(recent_resource_version_list)+1)[-1]
            else: 
                previous_resource_version = recent_resource_version_list[i]    

        return previous_resource_version

    def get_currently_deployed_model_version(self, model_version, return_latest_if_no_deployed=True, **kwargs): 
        model = self.run.models.get(model_version.parent)
        deployed_mv = None
        try: 
            #get promoted model version
            latest_promotion = model.latest_promotion()
            if latest_promotion: 
                deployed_mv_name = latest_promotion.model_version
                #if no promotion, system returns this shell object, so check for that
                if len(deployed_mv_name) > 0: 
                    deployed_mv = model.model_versions.get(deployed_mv_name) 
            #if not promotion exists, return prior mv
            if deployed_mv is None and return_latest_if_no_deployed:
                deployed_mv = self.get_prev_resource_version(model_version)
        except: 
            deployed_mv = None         
        return deployed_mv

    def get_prediction_job_model_version(self, prediction_job): 
        model_version_name = prediction_job.model_version 
        model_version = self.run.model_versions.get(model_version_name)
        return model_version
    
    def get_latest_model_resource(self, model, type): 
        resource = None 
        if type == "prediction_job": 
            resource = model.batch_predictions.list(default_sort_order="DESC")[0]
        return resource

    #dataset versions should be populated by process pipeline and passed in by runner
    #if that doesn't happen, we can look up the last model version and reuse those dataset versions
    #this is technically not foolproof, but it's unclear how best to try to establish these relationships
    #otherwise. We need something like default level model relationships, or getting them via a feature store, etc 
    def build_model_lineage(self, model_version, dataset_versions): 
        if len(dataset_versions) > 0: 
            for dv in dataset_versions: 
                dv.assignments.create(resource_name=model_version.name)
                #when supported, we should add assignments to other things we want to track: 
                #<code>.py, requirements.txt, docker containers, etc
                #for now we an just log this as metadata/artifacts
            
            # turn datasets into a dict and save
            # we're doing this so from mv we can reference all datasets used from the model. 
            # This is useful when we want to be able to retrieve training data used later on
            dataset_arr = [dv.name for dv in dataset_versions]
            dataset_dict = {x:y for x,y in zip(range(len(dataset_arr)),dataset_arr)}
            self.log_metadata(model_version, id="dataset_versions", data=dataset_dict)
        else: 
            prev_model_version = self.get_prev_resource_version(model_version)
            dataset_verisons = self.get_metadata(prev_model_version, id="dataset_versions") or {}
            for v in dataset_verisons.keys(): 
                dataset_version = self.get_resource(v, type="dataset_version")
                dv.assignments.create(resource_name=model_version.name)


    def get_resource_id(self, resource, *args, **kwargs): 
        return resource.id

    def get_parent_id(self, resource, *args, **kwargs):  
        return resource.parent.split("/")[-1]

    def register_vc_resource(self, resource, vc_info, key=None, file_type="csv", additional_metadata = {}, **kwargs): 
        #check if a git commit is present. if so then we know we did an external save
        if "hexsha" in vc_info.keys(): 
            uri = vc_info.get("uri")
            hexsha = vc_info.get("hexsha")

            artifact = self.log_artifact(resource, id = key, url=uri, mime_type=file_type, external=True)
            self.log_metadata(resource, id="git_hexsha_%s" %key, data={"hexsha": hexsha})
        
        if len(additional_metadata) > 0: 
            for k,v in additional_metadata.items(): 
                self.log_metadata(resource, id = k, data=v)

    def get_vc_info(self, resource, key = "git_hexsha"): 
        data = None
        metadata = self.get_metadata(resource, key)
        if metadata is not None: 
            data = metadata.get("hexsha")
        return data

    def log_data_profile(self, resource, file_path, profile, profiler_class = None): 
        # in the future we may want to try to map the profile to Continual's internal profile type. 
        # we would do that via the profiler class and definining a mapping between objects 
        # but for now let's just register as an artifact and then we can view it in the UI

        if file_path is not None: # then we created a profile externally from Continual 
            artifact = self.log_artifact(resource, id="data-profile", path = file_path, external=False)
            return artifact
        elif profiler_class == "ContinualDataProfiler": 
            profile.get("dataframes")[0].info()
            data_profile = resource.data_profiles.create(**profile) #in this case, profile is a dict of arguments for the log_data_profile 
            self.log("Created Continual data profile: %s" %data_profile.name)
            return data_profile 

    def log_checks(self, resource, file_path, report, checker_class = None, type="data"): 
        # in the future we probably want to parse the report based on the checker_class
        # and then recreate checks in Continual. but for now we'll just register as an artifact

        if file_path is not None: # then we created a profile externally from Continual 
            artifact = self.log_artifact(resource, id="%s-checks-report" %type, path = file_path, external=False)
            return artifact

    def get_data_profile(self, resource, id):
        return self.get_artifact(resource, id) 

    def log_data_comparison(self, resource, file_path, report, profiler_class = None): 
        # in the future we may want to try to map the profile to Continual's internal metrics types. 
        # we would do that via the profiler class and definining a mapping between objects 
        # but for now let's just register as an artifact and then we can view it in the UI

        if file_path is not None: # then we created a profile externally from Continual 
            artifact = self.log_artifact(resource, id="data-comparison", path = file_path, external=False)
            return artifact
        elif profiler_class == "ContinualDataProfiler": 
            #not implemented
            return None

    def get_winning_experiment(self, model_version):
        winning_exp_id = self.metadata_tracker.get_metadata(model_version, "winning_experiment_id").get("winning_experiment_id")
        experiment = self.metadata_tracker.get_resource(winning_exp_id, type="experiment", parent=model_version)
        return experiment

    def stop(self):
        self.run.complete()



from lolpop.component.base_component import BaseComponent

class BaseMetadataTracker(BaseComponent): 

    url: str = "https://replace.me"

    def log_artifact(self, resource, *args, **kwargs): 
        pass 

    def get_artifact(self, resource, id, *args, **kwargs):
        pass

    def log_tag(self, resource, *args, **kwargs):
        pass
    
    def get_tag(self, resource, key, *args, **kwargs):
        pass

    def log_metadata(self, resource, *args, **kwargs):
        pass 

    def get_metadata(self, resource, key, *args, **kwargs):
        pass

    def create_resource(self, type, *args, **kwargs):
        pass

    def get_resource(self, id, type, *args, **kwargs):
        pass

    def update_resource(self, resource, updates, *args, **kwargs):
        pass

    def get_prev_resource_version(self, resource_version, *args, **kwargs):
        pass 

    def get_currently_deployed_model_version(self, model_version, *args, **kwargs): 
        pass

    def get_prediction_job_model_version(self, prediction_job, *args, **kwargs): 
        pass 

    def get_latest_model_resource(self, model, type, *args, **kwargs): 
        pass 

    def get_winning_experiment(self, model_version, *args, **kwargs): 
        pass 

    def build_model_lineage(self, model_version, dataset_versions, *args, **kwargs): 
        pass 

    def get_resource_id(self, resource, *args, **kwargs):
        pass 

    def get_parent_id(self, resource, *args, **kwargs): 
        pass 

    def register_vc_resource(self, resource, vc_info, *args, **kwargs):
        pass

    def get_vc_info(self, resource, *args, **kwargs): 
        pass

    def log_data_profile(self, resource, *args, **kwargs):
        pass 
    
    def get_data_profile(self, resource, id, *args, **kwargs):
        pass 

    def log_checks(self, resource, *args, **kwargs):
        pass 
    
    def get_data_checks(self, resource, id, *args, **kwargs):
        pass 

    def log_data_comparison(self, resource, report, *args, **kwargs): 
        pass 

    def stop(self): 
        pass
    


from lolpop.runner.base_runner import BaseRunner
from lolpop.utils import common_utils as utils


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class MetaflowClassificationRunner(BaseRunner):

    __REQUIRED_CONF__ = {
        "pipelines": ["process", "train", "deploy", "predict"],
        "components": ["metadata_tracker", "metrics_tracker", "resource_version_control"],
        "config": ["train_data", "eval_data", "prediction_data", "model_target", "drop_columns"]
    }

    def __init__(self, problem_type="classification", *args, **kwargs):
        super().__init__(problem_type=problem_type, *args, **kwargs)

    def process_data(self, source = "train"):
        #run the metaflow process pipeline
        source_data_name = "%s_data" % source
        source_data = self._get_config(source_data_name)
        self.process.run(source_data_name, source_data)

        #retrieve artifacts from metaflow pipeline
        data, dataset_version = self.process.get_artifacts(["data", "dataset_version"])

        #return data
        return data, dataset_version

    def train_model(self, data, dataset_version=None):

        if data is None:
            if dataset_version is not None:
                data = self.resource_version_control.get_data(
                    dataset_version, self.metadata_tracker.get_vc_info(dataset_version))
            else:
                self.notify(
                    "No data provided. Can't train a model", level="ERROR")
                raise Exception("No data provided. Can't train a model.")

        #run the metaflow train pipeline
        self.train.run(data)

        #retrieve artifacts from metaflow pipeline
        model_version, model, is_new_model_better = self.train.get_artifacts(
            ["model_version", "model", "is_new_model_better"])

        return model_version, model, is_new_model_better

    def deploy_model(self, model_version, model):
        #run the metaflow deploy pipeline
        self.deploy.run(model, model_version)

        #retrieve artifacts from metaflow pipeline
        deployment = self.deploy.get_artifacts(
            ["deployment"])

        return deployment

    def predict_data(self, model_version, model, data, dataset_version):
        if model is None: 
            if model_version is not None: 
                experiment = self.metadata_tracker.get_winning_experiment(model_version)
                model_obj = self.resource_version_control.get_model(experiment)
                model = self.metadata_tracker.load_model(model_obj, model_version, model)

        #run the metaflow predict pipeline
        self.predict.run(model, model_version, data, dataset_version)

        #retrieve artifacts
        data, prediction_job = self.predict.get_artifacts(["data", "prediction_job"])

        return data, prediction_job

    #todo: move this into a new pipeline type: "evaluate"?
    def evaluate_ground_truth(self, prediction_job=None): 
        #if prediction job isn't known, get the most recent job
        if prediction_job == None: 
            model = self.metadata_tracker.get_resource(self._get_config("model_name"), type="model")
            prediction_job = self.metadata_tracker.get_latest_model_resource(model, type="prediction_job")
        
        #get prediciton data 
        vc_info = self.metadata_tracker.get_vc_info(prediction_job)
        prediction_data = self.resource_version_control.get_data(prediction_job, vc_info) 
        
        #get *current* training data
        train_data = self.process.transform_data(self._get_config("train_data"))
        
        #let's find common index values that exist in both the training dataset and the prediction set
        index = self._get_config("model_index")
        common_indices = train_data.merge(prediction_data, on=index, how="inner")[index]

        #if there are any in common, then we can check the ground truth 
        if len(common_indices) > 0: 
            #filter out any values not in the common index and then sort data so they match indices 
            train_data_filtered = train_data[train_data[index].isin(common_indices)]
            prediction_data_filtered = prediction_data[prediction_data[index].isin(common_indices)]
            train_data_sorted = train_data_filtered.sort_values(by=index).reset_index(drop=True)
            prediction_data_sorted = prediction_data_filtered.sort_values(by=index).reset_index(drop=True)
            ground_truth = {"y_train": train_data_sorted[self._get_config("model_target")]}
            predictions = {"train": prediction_data_sorted["predictions"]}
          
            #get model object and calculate metrics
            model_version = self.metadata_tracker.get_prediction_job_model_version(prediction_job)
            experiment = self.metadata_tracker.get_winning_experiment(model_version)
            model_obj = self.resource_version_control.get_model(experiment)
            model = self.metadata_tracker.load_model(model_obj, model_version, None)
            metrics_val = model.calculate_metrics(ground_truth, predictions, self.train._get_config("metrics"))

            #log stuff
            self.metrics_tracker.log_metrics(prediction_job, metrics_val, self.train._get_config("perf_metric"))
            self.metrics_tracker.log_metric(prediction_job, "num_ground_truth_observations", train_data_sorted.shape[0])
        else: #nothing in prediction has a ground truth yet 
            self.notify("Current training data has no overlap with the prediction job %s" %self.metadata_tracker.get_resource_id(prediction_job), level = "WARNING")

    def stop(self):
        self.metadata_tracker.stop()
        pass

    def build_all(self):
        data, dataset_version = self.process_data()
        model_version, model, is_new_model_better = self.train_model(data, dataset_version)
        if is_new_model_better: 
            deployment = self.deploy_model(model_version)
        eval_data, eval_dataset_version = self.process_data(source="eval")
        data, prediction_job = self.predict_data(model_version,model, eval_data, eval_dataset_version)
        self.evaluate_ground_truth(prediction_job)
    



from lolpop.component.data_connector.base_data_connector import BaseDataConnector
from lolpop.utils import common_utils as utils

import pandas as pd
import duckdb

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class DuckDBDataConnector(BaseDataConnector):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.path = self._get_config("duckdb_path")

    def get_data(self, table, sql=None, *args, **kwargs):
        if sql is None:
            if table is not None:
                sql = "SELECT * FROM %s" % table
            else:
                raise Exception(
                    "Table or SQL command must be provided to DuckDBDataConnector in order to retrieve data.")

        data = self._load_data(sql, self.path)

        return data

    def save_data(self, data, table, *args, **kwargs):
        #check if table already exists
        tables = self.get_data(None, sql="SHOW TABLES")

        #if table exists, see if we've added new columns or deleted columns
        table_exists = tables[tables["name"]==table].shape[0] > 0
        if table_exists:
           table_schema = self.get_data(table)
           new_features = [x for x in data.columns if x.upper(
           ) not in table_schema.columns.str.upper()]
           deleted_features = [
               x for x in table_schema.columns if x.upper() not in data.columns.str.upper()]

           #if we have new columns, we need to alter the table definiton
           if len(new_features) > 0:
               sql = "ALTER TABLE %s " % (table)
               for col_name in new_features:
                    col_type = data.dtypes[col_name]
                    duckdb_type = self._map_pandas_col_type_to_duckdb_type(col_type)
                    sql = sql + ("ADD COLUMN %s %s" %
                                 (col_name.upper(), duckdb_type))
               self.get_data(None, sql=sql)

           #if we've deleted any features, then just add blank rows to the dataframe
           if len(deleted_features) > 0:
               for feature in deleted_features:
                   data[feature] = None

        #now we can save our data
        self._save_data(data, table, self.path, table_exists)

     #load data into df
    @classmethod
    def _load_data(self, sql, path):
        result = pd.DataFrame()
        with duckdb.connect(database = path) as con: 
            res = con.sql(sql)
            result = res.df() 
        return result

    def _save_data(self, data, table_name, path, table_exists):
        if data is not None and len(data)>0: 
            with duckdb.connect(database = path) as con: 
                if not table_exists: 
                    con.sql("CREATE TABLE %s as SELECT * from data" %table_name)
                else: 
                    con.sql("INSERT INTO %s as SELECT * from data" % table_name)


    def _map_pandas_col_type_to_duckdb_type(self, col_type):
        """_summary_

        Args:
            col_type (_type_): _description_

        Returns:
            _type_: _description_
        """
        if col_type.kind == 'M':
            # datetime and timestamp columns in pandas are converted to datetime64[ns] dtype, which corresponds to 'TIMESTAMP'
            column_type = 'TIMESTAMP'
        elif col_type.kind == 'm':
            # timedelta columns in pandas are converted to timedelta64[ns] dtype, which corresponds to 'INTERVAL'
            column_type = 'INTERVAL'
        elif col_type.kind == 'b':
            # boolean columns in pandas are converted to 'bool' dtype, which corresponds to 'boolean'
            column_type = 'BOOLEAN'
        elif col_type.kind == 'i': 
            column_type = 'BIGINT'
        elif col_type.kind == 'u':
            column_type = 'UBIGINT'
        elif col_type.kind == 'f' or col_type.kind == 'c':
            # float columns in pandas are converted to 'float64' dtype, which corresponds to 'float'
            column_type = 'FLOAT'
        else:
            # All other column types in pandas are assumed to be string columns, which correspond to 'varchar' in Snowflake
            column_type = 'VARCHAR'
        return column_type



from lolpop.component.base_component import BaseComponent

class BaseDataTransformer(BaseComponent): 

    def get_data(self, source_table_name, *args, **kwargs): 
        pass 

    def transform(self, data, *args, **kwargs):
        pass

    def save_data(self, data, *args, **kwargs):
        pass


from lolpop.utils import common_utils as utils
from inspect import currentframe
class BaseRunner: 

    __REQUIRED_CONF__ = {
        "pipelines": [], 
        "components": ["metadata_tracker"], 
        "config": []
    }
    __DEFAULT_CONF__ = {
        "config" : {}
    }

    suppress_logger = False 
    suppress_notifier = False

    def __init__(self, conf={}, problem_type = "unspecified_problem_type", plugin_paths=[], 
                 skip_config_validation=False, *args, **kwargs):
        #handle configuration 
        self.name = type(self).__name__
        self.type = self.__module__.split(".")[-2]
        self.integration_type = self.__module__.split(".")[-1]
        conf = utils.get_conf(conf)
        conf = utils.resolve_conf_variables(conf)
        conf = utils.copy_config_into(conf, self.__DEFAULT_CONF__) 
        if not skip_config_validation:
            conf = self._validate_conf(conf)
        self.config = conf.get("config", {})
        self.problem_type = conf.get("problem_type", problem_type)

        #handle plugins
        if len(plugin_paths) == 0: 
            plugin_paths=conf.get("plugin_paths",[])
        file_path = None
        if hasattr(self, "__file_path__"): 
            file_path = self.__file_path__
        plugin_mods = utils.get_plugin_mods(self, plugin_paths, file_path)
        self.plugin_mods = plugin_mods

        #config defines all pipelines in `pipelines`
        #format is 
        ## pipelines: 
        ##   <pipeline>: <class> 
        ## components: 
        ##    <component>: <class>
        # then you need to also have a directory pipeline/component with all pipelines/components defined in subdirectories
        # later on in your yaml you need to specify configuration for the pipline via 
        ## <pipeline>: 
        ##    <key>: <value>
        # components can be defined at the runner or pipeline level and this determines the component scope.  
        # components defined at the runner level are passed down to all pipelines and components, 
        # so that they are globally acessible, whereas a component defined at the pipeline level are only accessible
        # in that pipeline and components defined at that pipeline. 
        # componenets in the right scope should know about each other but pipelines act independently (for now, at least)

        #handle default components: logger, notifier, metadata_tracker
        runner_components = {}
        runner_components = utils.set_up_default_components(self, conf, self.config,
                                                            plugin_mods=plugin_mods, 
                                                            skip_config_validation=skip_config_validation,
                                                            components = runner_components)

        #build all other components
        for component in conf.get("components",{}).keys(): 
            #ignore logger and metadata_tracker since we have already set those up
            if component != "logger" and component !="metadata_tracker" and component !="notifier": 
                obj = utils.register_component_class(self, conf, component, runner_conf=self.config, parent_process=self.name,
                                                     problem_type=self.problem_type, dependent_components=runner_components, 
                                                     plugin_mods=plugin_mods, skip_config_validation=skip_config_validation)
                if obj is not None: 
                    self.log("Loaded class %s into component %s" %(type(getattr(self, component)).__name__, component))
                    runner_components[component] = obj
                else: 
                    raise Exception("Unable to load class for component %s" %component)

        #now that all component classes are built, we want to update all components so that they know about each other. 
        # there is probably a more elegant way to do this
        for obj in runner_components.values(): 
            obj._update_components(components = runner_components)

        #build all pipelines 
        for pipeline in conf.get("pipelines",{}).keys(): 
            utils.register_pipeline_class(self, conf, pipeline, runner_conf=self.config, parent_process=self.name,
                                          problem_type=self.problem_type, dependent_components=runner_components, 
                                          plugin_mods=plugin_mods, skip_config_validation=skip_config_validation)
            self.log("Loaded class %s into pipeline %s" %(type(getattr(self, pipeline)).__name__, pipeline))

    def _validate_conf(self, conf):
        missing, total_missing = utils.validate_conf(conf, self.__REQUIRED_CONF__, conf.get("components"))
        if total_missing>0: 
            #logger is not necessarily set up yet
            #self.logger.log("Missing the following from runner configuration: %s" %missing)
            raise Exception("Missing the following from runner configuration: %s" %missing) 
        return conf

    def log(self, msg, level="INFO", **kwargs): 
        if not self.suppress_logger: 
            self.logger.log(msg, level, process_name=self.name,
                            line_num=currentframe().f_back.f_lineno, **kwargs)

    def notify(self, msg, level="ERROR"): 
        if not self.suppress_notifier: 
            self.notifier.notify(msg, level)
            self.log("Notification Sent: %s" %msg, level)

    #helper function for lookup up config key
    def _get_config(self, key, default_value=None):
        key = key.lower()
        value = utils.lower_conf(self.config).get(key, None)
        return value
        
    def _set_config(self, key, value): 
        key = key.lower()
        self.config[key]=value